import sys
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()

from aiida.orm import load_node
from aiida.common.exceptions import InputValidationError,ValidationError
from aiida.orm.data.upf import get_pseudos_from_structure
from aiida.common.datastructures import calc_states
from collections import defaultdict
from aiida.orm.utils import DataFactory, CalculationFactory
from aiida.orm.data.base import Float, Str, NumericType, BaseType 
from aiida.work.workchain import WorkChain, while_ , Outputs
from aiida.work.workchain import ToContext as ResultToContext
from aiida.work.run import legacy_workflow
from aiida.work.run import run, submit
from aiida.common.links import LinkType
from aiida.orm.data.remote import RemoteData 
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida_yambo.calculations.gw  import YamboCalculation
from aiida_yambo.workflows.yambo_utils import generate_yambo_input_params, reduce_parallelism 
from aiida_quantumespresso.calculations.pw import PwCalculation


ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")

class YamboRestartWf(WorkChain):
    """This module interacts directly with the yambo plugin to submit calculations

    This module submits calculations using the yambo plugin, and manages them, including
    restarting the calculation in case of:
    1. Memory problems ( will reduce MPI parallelism before resubmitting)
    2. Queue time exhaustions (will increase time by a fraction before resubmitting)
    3. Parallelism errors (will reduce the MPI the parallelism before resbmitting)
    4. Errors originating from a few select unphysical input parameters like too low bands.
    """

    @classmethod
    def define(cls, spec):
        """Workflow input parameters and the spec definition

        This function has a list of inputs that this workflow accepts, as well as the
        high level workflow iteration logic
        
        Keyword arguments:
        precode -- the P2Y code  (required) 
        yambocode -- the yambp code  (required) 
        calculation_set -- scheduler settings  (required) 
        settings -- code settings  (required) 
        parent_folder -- parent NSCF/P2Y/YAMBO calculation  (required) 
        parameters -- yambo parameter  (required)  
        restart_options -- the P2Y code  (optional) 
        """
        super(YamboRestartWf, cls).define(spec)
        spec.input("precode", valid_type=Str)
        spec.input("yambocode", valid_type=Str)
        spec.input("calculation_set", valid_type=ParameterData)
        spec.input("settings", valid_type=ParameterData)
        spec.input("parent_folder", valid_type=RemoteData)
        spec.input("parameters", valid_type=ParameterData)
        spec.input("restart_options", valid_type=ParameterData, required=False)
        spec.outline(
            cls.yambobegin,
            while_(cls.yambo_should_restart)(
                cls.yambo_restart,
                cls.interstep,
            ),
            cls.report_wf
        )
        spec.dynamic_output()

    def yambobegin(self):
        """Submits a calculation  using the  yambo plugin.

        This function takes inputs provided and using the YamboCalculation class
        submits a calculation.
        """
        self.ctx.yambo_pks = []
        self.ctx.yambo_nodes = []
        self.ctx.restart = 0 
        # run YamboCalculation
        if not isinstance(self.inputs.parent_folder, RemoteData):
            raise InputValidationError("parent_calc_folder must be of"
                                       " type RemoteData")        
        self.ctx.last = ''
        self.ctx.parameters = self.inputs.parameters
        self.ctx.calculation_set = self.inputs.calculation_set
        new_settings = self.inputs.settings.get_dict()
        parent_calc = self.inputs.parent_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
        yambo_parent = isinstance(parent_calc, YamboCalculation)
        if 'INITIALISE' not in new_settings.keys() and not yambo_parent:
            new_settings['INITIALISE'] = True
            self.ctx.last = 'INITIALISE'
        try:
            restart_options = self.inputs.restart_options
            try:
                max_restarts = restart_options.get_dict()['max_restarts']
            except KeyError:
                max_restarts = 5
        except AttributeError:
            restart_options = None
            max_restarts = 5
        self.ctx.max_restarts = max_restarts
        inputs = generate_yambo_input_params(
            self.inputs.precode,self.inputs.yambocode,
            self.inputs.parent_folder, self.ctx.parameters, self.ctx.calculation_set, ParameterData(dict=new_settings) )
        future = self.run_yambo(inputs)
        self.report("workflow start, submitted  {}".format(future.pid))
        return  ResultToContext(yambo= future)

    def interstep(self):
        # convenience function that stores output of resolved future  before the next loop if any, 
        # of no loop will run, it is still useful for the report_wf to find the resolved future's node
        # in the context. 
        self.ctx.yambo_nodes.append(self.ctx.yambo)

    def yambo_should_restart(self):
        """This function encodes the logic to restart calculations from failures

        This function supports detecting failed/incomplete yambo runs, taking corrective
        action and resubmitting automatically. These classes of errors are taken care of:
        1. Memory probelems
        2. Parallelism problems
        3. Some input inconsistency problems (too low bands)
        4. Queue time exhaustion.
        The calculation is restarted upto a maximum number of 4 retries
        """
        self.report("Checking if yambo restart is needed")
        if self.ctx.restart >= self.ctx.max_restarts:
            self.report("I will not restart: maximum restarts reached: {}".format(self.ctx.max_restarts))
            return False

        self.report("I can restart (# {}), max restarts ({}) not reached yet".format(self.ctx.restart, self.ctx.max_restarts))
        calc = load_node(self.ctx.yambo_pks[-1])
        if self.ctx.last == 'INITIALISE':
            return True

        if calc.get_state() == calc_states.SUBMISSIONFAILED:
            self.report("I will not resubmit calc pk: {}, submission failed: {}, check the log or you settings ".format(calc.pk ,calc.get_state() ))
            return False

        max_input_seconds = self.ctx.calculation_set.get_dict()['max_wallclock_seconds']

        last_time = 30 # seconds default value:
        try:
            last_time = calc.get_outputs_dict()['output_parameters'].get_dict()['last_time']  
        except Exception:
            pass  # Likely no logs were produced 
 
        if calc.get_state() == calc_states.FAILED and (float(max_input_seconds)-float(last_time))/float(max_input_seconds)*100.0 < 1:   
            max_input_seconds = int( max_input_seconds * 1.3) # 30% increase
            calculation_set = self.ctx.calculation_set.get_dict() 
            calculation_set['max_wallclock_seconds'] = max_input_seconds
            self.ctx.calculation_set = DataFactory('parameter')(dict=calculation_set) 
            self.report("Failed calculation, likely queue time exhaustion, restarting with new max_input_seconds = {}".format(
                        max_input_seconds ))
            return True

        if calc.get_state() != calc_states.PARSINGFAILED and calc.get_state != calc_states.FINISHED : # special case for parallelization needed
            output_p = {}
            if 'output_parameters'  in  calc.get_outputs_dict(): # calc.get_outputs_dict()['output_parameters'].get_dict().keys() 
                output_p = calc.get_outputs_dict()['output_parameters'].get_dict()
            if 'para_error' in output_p.keys(): 
                if output_p['para_error'] == True:  # Change parallelism or add missing parallelism inputs
                    self.report(" parallelism error detected")
                    params = self.ctx.parameters.get_dict() 
                    X_all_q_CPU = params.pop('X_all_q_CPU','')
                    X_all_q_ROLEs =  params.pop('X_all_q_ROLEs','') 
                    SE_CPU = params.pop('SE_CPU','')
                    SE_ROLEs = params.pop('SE_ROLEs','')
                    calculation_set = self.ctx.calculation_set.get_dict()
                    params['X_all_q_CPU'],calculation_set =   reduce_parallelism('X_all_q_CPU', X_all_q_ROLEs,  X_all_q_CPU, calculation_set )
                    params['SE_CPU'], calculation_set=  reduce_parallelism('SE_CPU', SE_ROLEs,  SE_CPU,  calculation_set )
                    params["X_all_q_ROLEs"] = X_all_q_ROLEs
                    params["SE_ROLEs"]= SE_ROLEs
                    self.ctx.calculation_set = DataFactory('parameter')(dict=calculation_set)
                    self.ctx.parameters = DataFactory('parameter')(dict=params)
                    self.report("Calculation {} failed from a parallelism problem: {}".format(calc.pk,output_p['errors']) )
                    self.report("Old parallelism {}= {} , {} = {} ".format(
                                    X_all_q_ROLEs,X_all_q_CPU, SE_ROLEs, SE_CPU))
                    self.report("New parallelism {}={} , {} = {}".format(
                                    X_all_q_ROLEs, params['X_all_q_CPU'], SE_ROLEs,  params['SE_CPU'] ))
                    return True 
            if 'unphysical_input' in output_p.keys():
                if output_p['unphysical_input'] == True:
                    # this handles this type of error: "[ERROR][NetCDF] NetCDF: NC_UNLIMITED in the wrong index"
                    # we should reset the bands to a larger value, it may be too small. 
                    # this is a probable cause, and it may not be the real problem, but often is the cause.
                    self.report("the calculation failed due to a problematic input, defaulting to increasing bands")
                    params = self.ctx.parameters.get_dict()
                    bandX = params.pop('BndsRnXp', None)
                    bandG = params.pop('GbndRnge', None)
                    if bandX:
                        bandX = ( bandX[0], int(bandX[0]*1.5)) # 
                        params['BndsRnXp'] = bandX
                    if bandG:
                        bandG = ( bandG[0], int(bandG[0]*1.5)) # 
                        params['GbndRnge'] = bandG
                    self.ctx.parameters = DataFactory('parameter')(dict=params)
                    return True 
                   
            if 'errors' in output_p.keys() and calc.get_state() == calc_states.FAILED:
                if len(calc.get_outputs_dict()['output_parameters'].get_dict()['errors']) < 1:
                    # No errors, We  check for memory issues, indirectly
                    if 'last_memory_time' in calc.get_outputs_dict()['output_parameters'].get_dict().keys():
                        # check if the last alloc happened close to the end:
                        last_mem_time = calc.get_outputs_dict()['output_parameters'].get_dict()['last_memory_time']
                        if  abs(last_time - last_mem_time) < 3: # 3 seconds  selected arbitrarily,
                            # this is (based on a simple heuristic guess, a memory related problem)
                            # change the parallelization to account for this before continuing, warn user too.
                            params = self.ctx.parameters.get_dict() 
                            X_all_q_CPU = params.pop('X_all_q_CPU','')
                            X_all_q_ROLEs =  params.pop('X_all_q_ROLEs','') 
                            SE_CPU = params.pop('SE_CPU','')
                            SE_ROLEs = params.pop('SE_ROLEs','')
                            calculation_set = self.ctx.calculation_set.get_dict()
                            params['X_all_q_CPU'],calculation_set =   reduce_parallelism('X_all_q_CPU', X_all_q_ROLEs,  X_all_q_CPU, calculation_set )
                            params['SE_CPU'], calculation_set=  reduce_parallelism('SE_CPU', SE_ROLEs,  SE_CPU,  calculation_set )
                            params["X_all_q_ROLEs"] = X_all_q_ROLEs
                            params["SE_ROLEs"]= SE_ROLEs
                            self.ctx.calculation_set = DataFactory('parameter')(dict=calculation_set)
                            self.ctx.parameters = DataFactory('parameter')(dict=params)
                            self.report("Calculation  {} failed likely from memory issues")
                            self.report("Old parallelism {}= {} , {} = {} ".format(
                                                  X_all_q_ROLEs,X_all_q_CPU, SE_ROLEs, SE_CPU))
                            self.report("New parallelism selected {}={}, {} = {} ".format(
                                                  X_all_q_ROLEs, params['X_all_q_CPU'], SE_ROLEs,  params['SE_CPU'] ))
                            return True 
                        else:
                            pass
            
        if calc.get_state() == calc_states.SUBMISSIONFAILED\
                   or calc.get_state() == calc_states.FAILED\
                   or 'output_parameters' not in calc.get_outputs_dict():
            self.report("Calculation {} failed or did not genrerate outputs for unknow reason, restarting with no changes".format(calc.pk))
            return True
        return False

    def yambo_restart(self):
        """Submits a yambo calculation using the yambo plugin

        This function submits a calculation, usually this represents a 
        resubmission of a failed calculation, or a continuation from P2Y.
        """

        calc = load_node(self.ctx.yambo_pks[-1])
        if  calc.get_state() == calc_states.SUBMISSIONFAILED:
            calc = self.get_last_submitted(calc.pk)
            if not calc: 
                raise ValidationError("restart calculations can not start from"
                                       "calculations in SUBMISSIONFAILED state")
                return        
        #parameters = calc.get_inputs_dict()['parameters'].get_dict()
        
        parent_folder = calc.out.remote_folder
        new_settings = self.inputs.settings.get_dict()

        parent_calc = parent_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
        yambo_parent = isinstance(parent_calc, YamboCalculation)
        p2y_restart =  calc.get_state() != calc_states.FINISHED and calc.inp.settings.get_dict().pop('INITIALISE',None)==True 
            
        if not yambo_parent or p2y_restart == True: 
            new_settings['INITIALISE'] =  True
            self.ctx.last = 'INITIALISE'
            self.report(" restarting from: {}, a P2Y calculation ".format(self.ctx.yambo_pks[-1]))
        else:
            new_settings['INITIALISE'] =  False
            self.ctx.last = 'NOTINITIALISE'
            self.report(" restarting from: {},  a GW calculation ".format(self.ctx.yambo_pks[-1]))
        inputs = generate_yambo_input_params(
             self.inputs.precode,self.inputs.yambocode,
             parent_folder, self.ctx.parameters, self.ctx.calculation_set, ParameterData(dict=new_settings) )
        future = self.run_yambo(inputs)
        self.ctx.yambo_pks.append(future.pid )
        self.ctx.restart += 1
        self.report(" restarting from:{}  ".format(future.pid )) 
        return ResultToContext(yambo_restart= future)

    def run_yambo(self,inputs):
        """Call submit with the inputs  """
        YamboProcess = YamboCalculation.process()
        future =  submit(YamboProcess, **inputs)
        self.ctx.yambo_pks.append( future.pid )
        self.report(" submitted a calculation with pk: {} ".format(future.pid ))
        return future  # we can not  ReturnToContext since this fuction is not called from the outline 


    def get_last_submitted(self, pk):
        submited = False
        depth = 0
        while not submited and depth <4:
            calc = load_node(pk)
            if  calc.get_state() == calc_states.SUBMISSIONFAILED:
                pk = load_node(calc.inp.parent_calc_folder.inp.remote_folder.pk).pk
            else:
                submited = calc
            depth+=1
        return  submited

    def report_wf(self):
        """
        Output final quantities
        return information that may be used to figure out
        the status of the calculation.
        """
        from aiida.orm import DataFactory
        success = load_node(self.ctx.yambo_pks[-1]).get_state()== 'FINISHED' 
        self.out("gw", DataFactory('parameter')(dict={ "yambo_pk":   self.ctx.yambo_pks[-1],  "success": success }))
        self.out("yambo_remote_folder",  load_node(self.ctx.yambo_pks[-1]).out.remote_folder )
        self.report("workflow completed ")

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

from aiida.orm.data.folder import FolderData
from aiida.parsers.parser import Parser
from aiida.common.datastructures import calc_states
from aiida.parsers.exceptions import OutputParsingError
from aiida.common.exceptions import UniquenessError
from aiida.common.exceptions import ValidationError, ParsingError
import numpy
import copy
from aiida.orm.data.array import ArrayData
from aiida.orm.data.array.bands import BandsData
from aiida.orm.data.array.kpoints import KpointsData
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.structure import StructureData
from aiida.orm.utils import DataFactory, CalculationFactory
import glob, os, re
from aiida_yambo.parsers.ext_dep.yambofile  import  YamboFile
from aiida_yambo.parsers.ext_dep.yambofolder  import  YamboFolder
from aiida_yambo.calculations.gw import YamboCalculation
#PwCalculation = CalculationFactory('quantumespresso.pw')
from aiida_quantumespresso.calculations.pw import PwCalculation
from aiida_quantumespresso.calculations import  get_input_data_text,_lowercase_dict,_uppercase_dict

__copyright__ = u"Copyright (c), 2014-2015, École Polytechnique Fédérale de Lausanne (EPFL), Switzerland, Laboratory of Theory and Simulation of Materials (THEOS). All rights reserved."
__license__ = "Non-Commercial, End-User Software License Agreement, see LICENSE.txt file"
__version__ = "0.4.1"
__authors__ = "Michael Atambo, Antimo Marrazzo, Gianluca Prandini and the AiiDA team. The parser relies on the yamboparser module by Henrique Pereira Coutada Miranda."

class YamboParser(Parser):
    """This class is a wrapper class for the Parser class for Yambo calculators from yambopy.

    *IMPORTANT:* This plugin can parse netcdf files produced by yambo if the 
    python netcdf libraries are installed, otherwise they are ignored.
    Accepts data from yambopy's YamboFolder  as a list of YamboFile instances. 
    The instances of YamboFile have the following attributes:
   
    ::
      .data: A Dict, with k-points as keys and  in each futher a dict with obeservalbe:value pairs ie. { '1' : {'Eo': 5, 'B':1,..}, '15':{'Eo':5.55,'B': 30}... }
      .warnings:     list of strings, one warning  per string.
      .errors:       list of errors, one error per string.
      .memory        list of string, info on memory allocated and freed
      .max_memory    maximum memory allocated or freed during the run
      .last_memory   last memory allocated or freed during the run
      .last_memory_time   last point in time at which  memory was  allocated or freed
      .*_units       units (e.g. Gb or seconds)
      .wall_time     duration of the run (as parsed from the log file)
      .last_time     last time reported (as parsed from the log file)
      .kpoints: When non empty is a Dict of kpoint_index: kpoint_triplet values i.e.                  { '1':[0,0,0], '5':[0.5,0.0,5] .. }
      .type:   type of file according to YamboFile types include: 
      1. 'report'    : 'r-..' report files
      2. 'output_gw'  : 'o-...qp': quasiparticle output file   ...           .. etc
      N. 'unknown' : when YamboFile was unable to deduce what type of file
      .timing: list of timing info.  

    Saved data:

    o-..qp : ArrayData is stored in a format similar to the internal yambo db format (two arrays):
             [[E_o,E-E_o,S_c],[...]]  and
             [[ik,ib,isp],...]
             First is the observables, and the second array contains the kpoint index, band index
             and spin index if spin polarized else 0. BandsData can not be used as the k-point triplets
             are not available in the o-.qp file.

    r-..    : BandsData is stored with the proper list of K-points, bands_labels. 

    """
    
    def __init__(self,calculation):
        """Initialize the instance of YamboParser"""
        from aiida.common import aiidalogger
        self._logger = aiidalogger.getChild('parser').getChild(
            self.__class__.__name__)
        # check for valid input
        if not isinstance(calculation, YamboCalculation):
            raise OutputParsingError("Input calculation must be a YamboCalculation")
        self._calc = calculation
        
        self._eels_array_linkname = 'array_eels'
        self._eps_array_linkname = 'array_eps'
        self._alpha_array_linkname = 'array_alpha'
        self._qp_array_linkname = 'array_qp'
        self._ndb_linkname = 'array_ndb'
        self._ndb_QP_linkname = 'array_ndb_QP'
        self._ndb_HF_linkname = 'array_ndb_HFlocXC'
        self._lifetime_bands_linkname = 'bands_lifetime'
        self._quasiparticle_bands_linkname = 'bands_quasiparticle'
        self._parameter_linkname = 'output_parameters'
        super(YamboParser, self).__init__(calculation)
        
    def parse_from_calc(self):
        """Parses the datafolder, stores results.

        This parser for this code ...
        """
        from aiida.common.exceptions import InvalidOperation
        from aiida.common import aiidalogger
        from  aiida.utils.logger import get_dblogger_extra
        
        parserlogger = aiidalogger.getChild('yamboparser')
        logger_extra = get_dblogger_extra(self._calc)
        
        # suppose at the start that the job is unsuccessful, unless proven otherwise
        successful = False  
        
        # check whether the yambo calc was an initialisation (p2y) 
        try:
            settings_dict = self._calc.inp.settings.get_dict()
            settings_dict = _uppercase_dict(settings_dict, dict_name='settings')
        except AttributeError:
            settings_dict = {}
 
        initialise = settings_dict.pop('INITIALISE', None)

        # select the folder object
        out_folder = self._calc.get_retrieved_node()
        
        # check what is inside the folder
        list_of_files = out_folder.get_folder_list()
        
        try:
            input_params = self._calc.inp.parameters.get_dict()
        except AttributeError:
            if not initialise:
                raise ParsingError("Input parameters not found!")
            else:
                input_params = {}
        # retrieve the cell: if parent_calc is a YamboCalculation we must find the original PwCalculation
        # going back through the graph tree.
        parent_calc = self._calc.inp.parent_calc_folder.inp.remote_folder
        cell = {}
        if isinstance(parent_calc, YamboCalculation):
            has_found_cell = False
            while (not has_found_cell):
                try:
                    cell = parent_calc.inp.structure.cell
                    has_found_cell = True
                except AttributeError:
                    parent_calc = parent_calc.inp.parent_calc_folder.inp.remote_folder    
        elif isinstance(parent_calc, PwCalculation):
            cell = self._calc.inp.parent_calc_folder.inp.remote_folder.inp.structure.cell

        output_params = {'warnings':[], 'errors':[],'yambo_wrote': False}
        new_nodes_list= []
        ndbqp = {}
        ndbhf = {}
        try:                          
            results = YamboFolder(out_folder.get_abs_path())
        except Exception, e:
            success = False 
            raise ParsingError("Unexpected behavior of YamboFolder: %s"%e)
  
        for result in results.yambofiles:
            if results is None:
                continue
            if result.max_memory:
                output_params['max_memory'] = result.max_memory    # Gb
                output_params['max_memory_units'] = 'Gb'    # Gb
            if result.last_memory:
                output_params['last_memory'] = result.last_memory    # Gb
                output_params['last_memory_units'] = 'Gb'    # Gb
            if result.last_memory_time:
                output_params['last_memory_time'] = result.last_memory_time    # seconds
                output_params['last_memory_time_units'] = 'seconds'    #  seconds
            if result.wall_time:
                output_params['wall_time'] = result.last_time # seconds
                output_params['wall_time_units'] = 'seconds' # seconds
            if result.last_time:
                output_params['last_time'] = result.last_time # seconds
                output_params['last_time_units'] = 'seconds' # seconds
            if result.yambo_wrote:
                output_params['yambo_wrote'] = True # boolean
            if result.timing:
                output_params['timing'] = result.timing
            if result.warnings:
                output_params['warnings'].extend(result.warnings)
            if result.errors:
                for err in result.errors:
                   if 'STOP' in err:
                       successful = False
                       break
                   else:
                       output_params['errors'].extend(result.errors)
            if  hasattr(result, 'para_error'):
                if result.para_error == True:
                    output_params['para_error'] = True
                else:
                    output_params['para_error'] = False
            if  hasattr(result, 'game_over'):
                if result.game_over == True:
                    successful = True 
            if initialise:
                # we do not have game_over, but we do have P2Y completed. 
                if  hasattr(result, 'p2y_completed'):
                    result.p2y_completed = True
                    successful = True
                
            if 'eel' in result.filename:
                eels_array = self._aiida_array(result.data)
                new_nodes_list.append( (self._eels_array_linkname, eels_array) )
            elif 'eps' in result.filename:
                eps_array = self._aiida_array(result.data)
                new_nodes_list.append( (self._eps_array_linkname, eps_array) )
            elif 'alpha' in result.filename:
                alpha_array = self._aiida_array(result.data)
                new_nodes_list.append( (self._alpha_array_linkname, alpha_array) )

            elif 'ndb.QP' == result.filename:
                 ndbqp = copy.deepcopy( result.data)

            elif 'ndb.HF_and_locXC' == result.filename:
                 ndbhf = copy.deepcopy(result.data)

            elif 'gw0' in input_params:
                if self._aiida_bands_data(result.data,cell,result.kpoints):
                    arr = self._aiida_bands_data(result.data, cell,result.kpoints)
                    if  type(arr)==BandsData: # ArrayData is not BandsData, but BandsData is ArrayData
                        new_nodes_list.append((self._quasiparticle_bands_linkname, arr ))
                    if type(arr) == ArrayData: # 
                        new_nodes_list.append((self._qp_array_linkname,arr ))

            elif 'life' in input_params:
                if self._aiida_bands_data(result.data,cell,result.kpoints):
                    arr = self._aiida_bands_data(result.data, cell,result.kpoints)
                    if type(arr) == BandsData:
                         new_nodes_list.append( (self._alpha_array_linkname, arr ))
                    elif type(arr)==ArrayData:
                        new_nodes_list.append((self._alpha_array_linkname+'_', arr ))

            else: 
                if not initialise:
                    output_params['warnings'].extend('Parser output format is invalid')
                else:
                    pass
        # we store  all the information from the ndb.* files rather than in separate files
        # if possible, else we default to separate files.
        if ndbqp and ndbhf:# 
            new_nodes_list.append((self._ndb_linkname, self._sigma_c(ndbqp,ndbhf)))
        else:
            if ndbqp:
                 new_nodes_list.append((self._ndb_QP_linkname,self._aiida_ndb_qp(ndbqp)))
            if ndbhf:
                 new_nodes_list.append((self._ndb_HF_linkname,self._aiida_ndb_hf(ndbhf)))
             
        param = ParameterData(dict=output_params)
        new_nodes_list.append( (self._parameter_linkname, param) ) # output_parameters

        # successful=False -> Calc state = FAILED
        return successful, new_nodes_list

    def _aiida_array(self, data):
        arraydata = ArrayData()
        for ky in data.keys():
            arraydata.set_array(ky,  data[ky])
        return arraydata 

    def _aiida_bands_data(self, data,cell,kpoints_dict):
        if not data :
            return False 
        kpt_idx = sorted(data.keys()) #  list of kpoint indices 
        try:
            k_list = [ kpoints_dict[i] for i in kpt_idx ] # list of k-point triplet
        except KeyError: 
            # kpoint triplets are not present (true  for .qp and so on, can not use BandsData)
            # We use the internal Yambo Format  [ [Eo_1, Eo_2,... ], ...[So_1,So_2,] ] 
            #                                  QP_TABLE  [[ib_1,ik_1,isp_1]      ,[ib_n,ik_n,isp_n]]
            # Each entry in DATA has corresponding legend in QP_TABLE that defines its details
            # like   ib= Band index,  ik= kpoint index,  isp= spin polarization index. 
            #  Eo_1 =>  at ib_1, ik_1 isp_1.
            pdata  = ArrayData()
            QP_TABLE=[]
            ORD= []
            Eo =[] ; E_minus_Eo =[] ; So=[]; Z=[] ;  
            for ky in data.keys(): # kp == kpoint index as a string  1,2,..
                for ind in range(len( data[ky]['Band'])):
                    try:
                        Eo.append(data[ky]['Eo'][ind])
                    except KeyError:
                        pass
                    try:
                        E_minus_Eo.append(data[ky]['E-Eo'][ind])
                    except KeyError:
                        pass
                    try:
                        So.append(data[ky]['Sc|Eo'][ind])
                    except KeyError:
                        pass
                    try:
                        Z.append(data[ky]['Z'][ind])
                    except KeyError:
                        pass
                    ik = int(ky)
                    ib = data[ky]['Band'][ind]
                    isp=0
                    if 'Spin_Pol' in data[ky].keys():
                        isp = data[ky]['Spin_Pol'][ind]
                    QP_TABLE.append([ik, ib, isp])
            pdata.set_array('Eo', numpy.array(Eo))
            pdata.set_array('E_minus_Eo', numpy.array(E_minus_Eo))
            pdata.set_array('So', numpy.array(So))
            pdata.set_array('Z', numpy.array(Z))
            pdata.set_array('qp_table', numpy.array(QP_TABLE))
            return  pdata 
        quasiparticle_bands = BandsData()
        quasiparticle_bands.set_cell(cell)
        quasiparticle_bands.set_kpoints(k_list, cartesian=True)
        # labels will come from any of the keys in the nested  kp_point data,
        # there is a uniform set of observables for each k-point, ie Band, Eo, ...
        # ***FIXME BUG does not seem to handle spin polarizes at all when constructing bandsdata***
        bands_labels = [ legend for legend in sorted(data[data.keys()[0]].keys())]
        append_list =[[] for i in bands_labels]
        for kp in kpt_idx:
            for i in range(len(bands_labels)):
                append_list[i].append( data[kp][bands_labels[i]] )   
        generalised_bands = [ numpy.array(it) for it in append_list  ]
        quasiparticle_bands.set_bands(bands=generalised_bands,
              units='eV', labels=bands_labels)
        return quasiparticle_bands 

    def _aiida_ndb_qp(self, data ):
        """
        Save the data from ndb.QP to the db
        """
        pdata  = ArrayData()
        pdata.set_array('Eo', numpy.array(data['Eo']))
        pdata.set_array('E_minus_Eo', numpy.array(data['E-Eo']))
        pdata.set_array('Z', numpy.array(data['Z']))
        pdata.set_array('qp_table', numpy.array(data['qp_table']))
        try:
            pdata.set_array('So', numpy.array(data['So']))
        except KeyError:
            pass
        return pdata

    def _aiida_ndb_hf(self, data ):
        """Save the data from ndb.HF_and_locXC  

        """
        pdata  = ArrayData()
        pdata.set_array('Sx', numpy.array(data['Sx']))
        pdata.set_array('Vxc', numpy.array(data['Vxc']))
        return pdata

    def _sigma_c(self, ndbqp, ndbhf):
        """Calculate S_c if missing from  information parsed from the  ndb.*

         Sc = 1/Z[ E-Eo] -S_x + Vxc
        """
        Eo = numpy.array(ndbqp['E-Eo'])
        Z =   numpy.array(ndbqp['Z'])
        E_minus_Eo = numpy.array(ndbqp['E-Eo'])
        Sx =  numpy.array(ndbhf['Sx'])
        Vxc = numpy.array(ndbhf['Vxc'])
        try:
            Sc =  numpy.array(ndbqp['So'])
        except KeyError:   
            Sc = 1/Z*E_minus_Eo -Sx + Vxc 
        pdata = ArrayData()
        pdata.set_array('Eo', Eo)
        pdata.set_array('E_minus_Eo', E_minus_Eo)
        pdata.set_array('Z', Z)
        pdata.set_array('Sx', Sx)
        pdata.set_array('Sc', Sc)
        pdata.set_array('Vxc', Vxc)
        pdata.set_array('qp_table', numpy.array(ndbqp['qp_table']))
        return pdata     



# -*- coding: utf-8 -*-
"""
Plugin to create a Yambo input file.
"""
import os
from aiida.orm.calculation.job import JobCalculation
from aiida.common.exceptions import InputValidationError,ValidationError
from aiida.common.datastructures import CalcInfo
from aiida.common.datastructures import calc_states
from aiida_quantumespresso.calculations import  get_input_data_text,_lowercase_dict,_uppercase_dict
from aiida.common.exceptions import UniquenessError, InputValidationError
from aiida.common.utils import classproperty
from aiida.orm.data.parameter import ParameterData 
from aiida.orm.data.remote import RemoteData 
from aiida.orm.utils import DataFactory, CalculationFactory
from aiida.orm.code import Code
from aiida.common import aiidalogger
from aiida.common.links import LinkType
PwCalculation = CalculationFactory('quantumespresso.pw')

__copyright__ = u"Copyright (c), 2014-2015, École Polytechnique Fédérale de Lausanne (EPFL), Switzerland, Laboratory of Theory and Simulation of Materials (THEOS). All rights reserved."
__license__ = "Non-Commercial, End-User Software License Agreement, see LICENSE.txt file"
__version__ = "0.4.1"
__authors__ = "Gianluca Prandini, Antimo Marrazzo, Michael Atambo and the AiiDA team."
    
class YamboCalculation(JobCalculation):
    """
    Yambo code.
    For more information, refer to http://www.yambo-code.org/
    """

    def _init_internal_params(self):
        super(YamboCalculation, self)._init_internal_params()

        self._INPUT_FILE_NAME = 'aiida.in'
        
        #Maybe the output name is not necessary...
        self._OUTPUT_FILE_NAME = 'aiida'
    
        # Default output parser provided by AiiDA
        self._default_parser = 'yambo.yambo'
        
        # Default input and output files
        self._DEFAULT_INPUT_FILE = 'aiida.in'
        self._DEFAULT_OUTPUT_FILE = 'aiida.out'

        self._SCRATCH_FOLDER = 'SAVE'

        self._LOGOSTRING = """#                                                           
# Y88b    /   e           e    e      888~~\    ,88~-_      
#  Y88b  /   d8b         d8b  d8b     888   |  d888   \     
#   Y88b/   /Y88b       d888bdY88b    888 _/  88888    |    
#    Y8Y   /  Y88b     / Y88Y Y888b   888  \  88888    |    
#     Y   /____Y88b   /   YY   Y888b  888   |  Y888   /     
#    /   /      Y88b /          Y888b 888__/    `88_-~      
#                                                           
#             AIIDA input plugin.  YAMBO 4.x compatible
#               http://www.yambo-code.org                   
#
"""

    @classproperty
    def _use_methods(cls):
        """
        Additional use_* methods for the YamboCalculation class.
        """
        retdict = JobCalculation._use_methods
        retdict.update({
            "settings": {
               'valid_types': ParameterData,
               'additional_parameter': None,
               'linkname': 'settings',
               'docstring': "Use an additional node for special settings",
               },
            "parameters": {
               'valid_types': ParameterData,
               'additional_parameter': None,
               'linkname': 'parameters',
               'docstring': ("Use a node that specifies the input parameters "
                             "for the namelists"),
               },
            "parent_folder": {
               'valid_types': RemoteData,
               'additional_parameter': None,
               'linkname': 'parent_calc_folder',
               'docstring': ("Use a remote folder as parent folder (for "
                             "restarts and similar"),
               },
            "preprocessing_code": {'valid_types': Code,
                                   'additional_parameter': None,
                                   'linkname': 'preprocessing_code',
                                   'docstring': ("Use a preprocessing code for "
                                                 "starting yambo"),
               },
            "precode_parameters": {
                                   'valid_types': ParameterData,
                                   'additional_parameter': None,
                                   'linkname': 'precode_parameters',
                                   'docstring': ("Use a node that specifies the input parameters "
                                                 "for the yambo precode"),
               },
            })
        return retdict
    
    def _prepare_for_submission(self, tempfolder, inputdict):        
        """
        This is the routine to be called when you want to create
        the input files and related stuff with a plugin.
        
        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputdata_dict (with the Code(s)!)
        """
#        from aiida.common.utils import get_unique_filename, get_suggestion

        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []
        
        # Settings can be undefined, and defaults to an empty dictionary.
        # They will be used for any input that doen't fit elsewhere.
        settings = inputdict.pop(self.get_linkname('settings'),None)
        if settings is None:
            settings_dict = {}
        else:
            if not isinstance(settings,  ParameterData):
                raise InputValidationError("settings, if specified, must be of "
                                           "type ParameterData")
            # Settings converted to uppercase
            settings_dict = _uppercase_dict(settings.get_dict(),
                                            dict_name='settings')
        initialise = settings_dict.pop('INITIALISE', None)
        if initialise is not None:
            if not isinstance(initialise, bool):
                raise InputValidationError("INITIALISE must be "
                                       " a boolean")
        try:
            parameters = inputdict.pop(self.get_linkname('parameters'))
        except KeyError:
            if not initialise:
                raise InputValidationError("No parameters specified for this calculation")
            else:    
                pass
        if not initialise:
            if not isinstance(parameters, ParameterData):
                raise InputValidationError("parameters is not of type ParameterData")

        parent_calc_folder = inputdict.pop(self.get_linkname('parent_folder'),None)
        if parent_calc_folder is None:
            raise InputValidationError("No parent calculation found, it is needed to "
                                       "use Yambo")
        if not isinstance(parent_calc_folder, RemoteData):
            raise InputValidationError("parent_calc_folder must be of"
                                       " type RemoteData")
        
        ### !!!!!! ###
        main_code = inputdict.pop(self.get_linkname('code'),None)
        if main_code is None:
            raise InputValidationError("No input code found!")
        
        
        preproc_code =  inputdict.pop(self.get_linkname('preprocessing_code'),None)
        if preproc_code is not None:
            if not isinstance(preproc_code, Code):
                raise InputValidationError("preprocessing_code, if specified,"
                                           "must be of type Code")
        
        
        parent_calc = parent_calc_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
        yambo_parent = isinstance(parent_calc, YamboCalculation)
        
        # flags for yambo interfaces
        try:
            precode_parameters = inputdict.pop(self.get_linkname
                                               ('precode_parameters'))
        except KeyError:
            precode_parameters = ParameterData(dict={})
        if not isinstance(precode_parameters,ParameterData):
            raise InputValidationError('precode_parameters is not '
                                       'of type ParameterData')
        precode_param_dict = precode_parameters.get_dict()

        # check the precode parameters given in input
        input_cmdline = settings_dict.pop('CMDLINE', None)
        import re
        precode_params_list = []
        pattern = re.compile(r"(^\-)([a-zA-Z])")
        for key, value in precode_param_dict.iteritems():
            if re.search(pattern, key) is not None:
                if key == '-O' or key == '-H' or key == '-h' or key == '-F': 
                    raise InputValidationError("Precode flag {} is not allowed".format(str(key)))
                else:   
                    if  precode_param_dict[key] is True:
                        precode_params_list.append(str(key))   
                    elif  precode_param_dict[key] is False:
                        pass            
                    else:
                        precode_params_list.append('{}'.format(str(key)))
                        precode_params_list.append('{}'.format(str(value)))
            else:
                raise InputValidationError("Wrong format of precode_parameters")   
        # Adding manual cmdline input (e.g. for DB fragmentation) 
        if input_cmdline is not None: 
            precode_params_list = precode_params_list + input_cmdline
        
        # TODO: remotedata must be on the same computer 

        ##############################
        # END OF INITIAL INPUT CHECK #
        ##############################

        
        if not initialise:
        ###################################################
        # Prepare yambo input file
        ###################################################
        
            params_dict = parameters.get_dict()
            
            # extract boolean keys
            boolean_dict = { k:v for k,v in params_dict.iteritems() if isinstance(v,bool) }
            params_dict = { k:v for k,v in params_dict.iteritems() if k not in boolean_dict.keys() }
            
            # reorganize the dictionary and create a list of dictionaries with key, value and units
            parameters_list = []
            for k,v in params_dict.iteritems():
                
                if "_units" in k:
                    continue
                
                units_key = "{}_units".format(k)
                try:
                    units = params_dict[units_key]
                except KeyError:
                    units = None
                
                this_dict = {}
                this_dict['key'] = k
                this_dict['value'] = v
                this_dict['units'] = units
                
                parameters_list.append( this_dict )
            
            input_filename = tempfolder.get_abs_path(self._INPUT_FILE_NAME)
            
            ## create an empty folder for the Yambo scratch
            #tempfolder.get_subfolder(self._SCRATCH_FOLDER, create=True)
            
            with open(input_filename,'w') as infile:
                infile.write( self._LOGOSTRING)
                
                for k,v in boolean_dict.iteritems():
                    if v:
                        infile.write( "{}\n".format(k) )
                
    #         .format(key.lower(), value_string)   
                
                for this_dict in parameters_list:
                    key = this_dict['key']
                    value = this_dict['value']
                    units = this_dict['units']
                
                    if isinstance(value,(tuple,list)):
                        # write the input flags for the Drude term and for the parallelization options of vers. 4 
                        # (it can be implemented in a better way)
                        if key.startswith('DrudeW'):
                            value_string = " ( " + ",".join([str(_) for _ in value]) + " )"
                            the_string = "{} = {}".format(key, value_string)
                            the_string += " {}".format(units)
                            infile.write( the_string + "\n" )
                            continue
                        
                        if key == 'SE_CPU':
                            value_string = " \" " + " ".join([str(_) for _ in value]) + " \" "
                            the_string = "{} = {}".format(key, value_string)
                            infile.write( "SE_ROLEs = \" q qp b  \" " + "\n" )
                            infile.write( the_string + "\n" )
                            continue
                        
                        if key == 'X_all_q_CPU':
                            value_string = " \" " + " ".join([str(_) for _ in value]) + " \" "
                            the_string = "{} = {}".format(key, value_string)
                            infile.write( "X_all_q_ROLEs = \" q k c v  \" " + "\n" )
                            infile.write( the_string + "\n" )
                            continue
                        
                        if key == 'X_finite_q_CPU':
                            value_string = " \" " + " ".join([str(_) for _ in value]) + " \" "
                            the_string = "{} = {}".format(key, value_string)
                            infile.write( "X_finite_q_ROLEs = \" q k c v  \" " + "\n" )
                            infile.write( the_string + "\n" )
                            continue
                        
                        if key == 'X_q_0_CPU':
                            value_string = " \" " + " ".join([str(_) for _ in value]) + " \" "
                            the_string = "{} = {}".format(key, value_string)
                            infile.write( "X_q_0_ROLEs = \" k c v  \" " + "\n" )
                            infile.write( the_string + "\n" )
                            continue
                        
                        if key == 'QPkrange' or key == 'QPerange':
                            value_string = ''
                            for v in value:
                                value_string += " | ".join([str(_) for _ in v]) + " |\n"    
                            the_string = "% {}\n {}".format(key, value_string)    
                            the_string += "%"  
                            infile.write( the_string + "\n" )
                            continue    
                                                             
                        value_string = " | ".join([str(_) for _ in value]) + " |"
                        the_string = "% {}\n {}".format(key, value_string)
                        if units is not None:
                            the_string += " {}".format(units)
                        the_string += "\n%"
                            
                    else:
                        the_value = '"{}"'.format(value) if isinstance(value,basestring) else '{}'.format(value)
                        the_string = "{} = {}".format(key, the_value)
                        if units is not None:
                            the_string += " {}".format(units)
    
                    infile.write( the_string + "\n" )


        ############################################
        # set copy of the parent calculation
        ############################################
        
        parent_calcs = parent_calc_folder.get_inputs(link_type=LinkType.CREATE)
        if len(parent_calcs)>1:
            raise UniquenessError("More than one parent totalenergy calculation" 
                                  "has been found for parent_calc_folder {}".format(parent_calc_folder))
        if len(parent_calcs)==0:
            raise InputValidationError("No parent calculation associated with parent_folder {}".format(parent_calc_folder))
        parent_calc = parent_calcs[0]
        
        if yambo_parent:
            try:
                parent_settings = _uppercase_dict(parent_calc.inp.settings.get_dict(),
                                            dict_name='parent settings')
                parent_initialise = parent_settings['INITIALISE']
            except KeyError:
                parent_initialise = False
        
        if yambo_parent:
            remote_copy_list.append(
                                    (
                                     parent_calc_folder.get_computer().uuid,
                                     os.path.join(parent_calc_folder.get_remote_path(),"SAVE"),
                                     "SAVE/"
                                     )
                                    )
            if not parent_initialise:
                cancopy = False
                if parent_calc.get_state() == calc_states.FINISHED:
                    cancopy = True
                if 'yambo_wrote' in  parent_calc.get_outputs_dict()['output_parameters'].get_dict().keys():
                    if parent_calc.get_outputs_dict()['output_parameters'].get_dict()['yambo_wrote'] == True: 
                        cancopy = True 
                    if parent_calc.get_outputs_dict()['output_parameters'].get_dict()['yambo_wrote'] == False: 
                        cancopy = False 
                if cancopy:
                    remote_copy_list.append(
                                       (
                                         parent_calc_folder.get_computer().uuid,
                                         os.path.join(parent_calc_folder.get_remote_path(),"aiida"),
                                         "aiida/"
                                         )
                                        )
        else:    
            remote_copy_list.append(
                                    (
                                     parent_calc_folder.get_computer().uuid,
                                     os.path.join(parent_calc_folder.get_remote_path(),
                                                  PwCalculation._OUTPUT_SUBFOLDER,
                                                  "{}.save".format(parent_calc._PREFIX),"*" ),
                                     "."
                                     )
                                    )
        ############################################
        # set Calcinfo
        ############################################
                
        calcinfo = CalcInfo()
        
        calcinfo.uuid = self.uuid
        
        calcinfo.local_copy_list = []
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = []# remote_symlink_list
        #calcinfo.stdout_name = None # self._OUTPUT_FILE_NAME
        
        # Retrieve by default the output file and the xml file
        calcinfo.retrieve_list = []
        #calcinfo.retrieve_list.append(self._OUTPUT_FILE_NAME)
        calcinfo.retrieve_list.append('r*')
        calcinfo.retrieve_list.append('l*')
        calcinfo.retrieve_list.append('o*')        
        calcinfo.retrieve_list.append('LOG/l-*_CPU_1')        
        extra_retrieved = settings_dict.pop('ADDITIONAL_RETRIEVE_LIST', ['aiida/ndb.QP','aiida/ndb.HF_and_locXC'])
        for extra in extra_retrieved:
            calcinfo.retrieve_list.append( extra )
        
#        # Empty command line by default
#        cmdline_params = settings_dict.pop('CMDLINE', [])
#        calcinfo.cmdline_params = (list(cmdline_params)
#                                   + ["-F", self._INPUT_FILE_NAME, '-J', self._OUTPUT_FILE_NAME])
        
        from aiida.common.datastructures import code_run_modes, CodeInfo
        
        # c1 = interface dft codes and yambo (ex. p2y or a2y)
        c1 = CodeInfo()
        c1.withmpi = True
        c1.cmdline_params = precode_params_list
        
        # c2 = yambo initialization
        c2 = CodeInfo()
        c2.withmpi = True
        c2.cmdline_params = []
        c2.code_uuid = main_code.uuid
        
        # if the parent calculation is a yambo calculation skip the interface (c1) and the initialization (c2)
        if yambo_parent:
            c1 = None
            if not parent_initialise:
                c2 = None
        else:
            c1.cmdline_params = precode_params_list
            c1.code_uuid = preproc_code.uuid
            
        # c3 = yambo calculation
        c3 = CodeInfo()
        c3.withmpi = self.get_withmpi()
        c3.cmdline_params = ["-F", self._INPUT_FILE_NAME, '-J', self._OUTPUT_FILE_NAME]
        c3.code_uuid = main_code.uuid
        
        if initialise:
            c2 = None
            c3 = None
        
        #calcinfo.codes_info = [c1, c2, c3] if not yambo_parent else [c3]
        if yambo_parent:
            if not parent_initialise:
                calcinfo.codes_info=[c3]
            else:
                calcinfo.codes_info=[c2,c3]
        elif initialise:
            calcinfo.codes_info=[c1]
        else:
            calcinfo.codes_info=[c1, c2, c3] 
             
        calcinfo.codes_run_mode = code_run_modes.SERIAL


        if settings_dict:
            raise InputValidationError("The following keys have been found in "
                "the settings input node, but were not understood: {}".format(
                ",".join(settings_dict.keys())))

        return calcinfo
      
        
    def _check_valid_parent(self,calc):
        """
        Check that calc is a valid parent for a YamboCalculation.
        It can be a PwCalculation or a YamboCalculation.
        """

        try:
            if ( (not isinstance(calc,PwCalculation)) 
                  and (not isinstance(calc,YamboCalculation)) ):
                raise ValueError("Parent calculation must be a PwCalculation or a YamboCalculation")
                               
        except ImportError:
            if ( (not isinstance(calc,PwCalculation)) 
                            and (not isinstance(calc,YamboCalculation)) ):
                raise ValueError("Parent calculation must be a PwCalculation or a YamboCalculation")
                            
    
    def use_parent_calculation(self,calc):
        """
        Set the parent calculation of Yambo, 
        from which it will inherit the outputsubfolder.
        The link will be created from parent RemoteData to YamboCalculation 
        """
        from aiida.common.exceptions import NotExistent
        
        self._check_valid_parent(calc)
        
        remotedatas = calc.get_outputs(type=RemoteData)
        if not remotedatas:
            raise NotExistent("No output remotedata found in "
                                  "the parent")
        if len(remotedatas) != 1:
            raise UniquenessError("More than one output remotedata found in "
                                  "the parent")
        remotedata = remotedatas[0]
        
        self._set_parent_remotedata(remotedata)

    def _set_parent_remotedata(self,remotedata):
        """
        Used to set a parent remotefolder in the start of Yambo.
        """
        if not isinstance(remotedata,RemoteData):
            raise ValueError('remotedata must be a RemoteData')
        
        # complain if another remotedata is already found
        input_remote = self.get_inputs(node_type=RemoteData)
        if input_remote:
            raise ValidationError("Cannot set several parent calculation to a "
                                  "Yambo calculation")

        self.use_parent_folder(remotedata)



import sys
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()

from aiida.orm import load_node
from aiida.orm.data.upf import get_pseudos_from_structure
from aiida.common.exceptions import InputValidationError,ValidationError
from collections import defaultdict
from aiida.orm.utils import DataFactory, CalculationFactory
from aiida.orm.data.base import Float, Str, NumericType, BaseType, List, Bool
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida.work.run import run, submit
from aiida.work.workchain import WorkChain, while_, ToContext, Outputs
from aiida_yambo.calculations.gw  import YamboCalculation
from aiida.common.links import LinkType
from aiida_yambo.workflows.yambo_utils import default_step_size, update_parameter_field, set_default_qp_param,\
               default_pw_settings, set_default_pw_param, yambo_default_settings, default_qpkrange,\
               p2y_default_settings, is_converged
from aiida_yambo.workflows.yamborestart  import YamboRestartWf
from aiida_yambo.workflows.yambowf  import YamboWorkflow
from aiida.orm.data.remote import RemoteData
from aiida_quantumespresso.calculations.pw import PwCalculation
from aiida_yambo.calculations.gw  import YamboCalculation
import numpy as np 
from scipy.optimize import  curve_fit 


ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")
PwProcess = PwCalculation.process()
YamboProcess = YamboCalculation.process()


class YamboConvergenceWorkflow(WorkChain):
    """This class supports 1-D convergence of FFT, kpoints, Bands or W_cutoff, only one must be selected.

    This class computed the quasiparticle direct gap at Gamma, and updates the convergence paramteter
    until the change in at least three sequential calculations is within a set threshold.
    """

    @classmethod
    def define(cls, spec):
        """
        convergence_parameters = {'variable_to_converge':'bands' or 'W_cutoff' or 'kpoints' or 'FFT_cutoff',
                                    'start_value': 10,
                                    'step': 5,
                                    'max_value':100,
                                    'conv_tol': 0.1,
                                    'conv_window': 3 (optional),
                                    'loop_length': 4 (optional),
                                    }
        """
        super(YamboConvergenceWorkflow, cls).define(spec)
        spec.input("precode", valid_type=BaseType)
        spec.input("pwcode", valid_type=BaseType, required=False)
        spec.input("yambocode", valid_type=BaseType)
        spec.input("pseudo", valid_type=BaseType,required=True)
        spec.input("calculation_set_pw", valid_type=ParameterData, required=False)
        spec.input("calculation_set_p2y", valid_type=ParameterData,required=False)
        spec.input("calculation_set", valid_type=ParameterData )
        spec.input("parent_scf_folder", valid_type=RemoteData, required=False)
        spec.input("settings_p2y", valid_type=ParameterData, required=False, default = p2y_default_settings())
        spec.input("settings", valid_type=ParameterData, required=False, default = yambo_default_settings())
        spec.input("structure", valid_type=StructureData,required=False)
        spec.input("parent_nscf_folder", valid_type=RemoteData, required=False)
        spec.input("parameters_p2y", valid_type=ParameterData, required=False, default=set_default_qp_param()  )
        spec.input("parameters", valid_type=ParameterData, required=False  )
        spec.input("parameters_pw", valid_type=ParameterData, required=False  )
        spec.input("parameters_pw_nscf", valid_type=ParameterData, required=False  )
        spec.input("convergence_parameters", valid_type=ParameterData, required=True)

        spec.outline(
          cls.start,
          cls.iterate,
          while_(cls.is_not_converged)(
              cls.run_next_update,
              cls.iterate,
              ),
          cls.report_wf
        )
        spec.dynamic_output()



    def init_parameters(self,paging):
        """This function initializes the  settings and parameters needed for  a convergence calculations.

        This function will store all needed parameters each in a self.ctx variable, and generate from
        defaults optional parameters that are not provieded. It uses functions from `yambo_utils` to 
        generates the defaults. This function is only called once,  after which the only update parameter
        is called.
        """
        convergence_parameters_dict = self.inputs.convergence_parameters.get_dict()
        if 'calculation_set_pw' not in self.inputs.keys():
            self.inputs.calculation_set_pw = DataFactory('parameter')(dict=self.inputs.calculation_set.get_dict())

        if 'calculation_set_p2y' not in self.inputs.keys():
            main_set = self.inputs.calculation_set.get_dict()
            main_set['resources'] = {"num_machines": 1,"num_mpiprocs_per_machine":  1}
            self.ctx.calculation_set_p2y = DataFactory('parameter')(dict=main_set)
        else:
            self.ctx.calculation_set_p2y =  self.inputs.calculation_set_p2y 

        if 'parameters' not in self.inputs.keys():
            self.ctx.parameters = set_default_qp_param()
        else:
            self.ctx.parameters = self.inputs.parameters

        if 'parent_scf_folder' in  self.inputs.keys(): 
            self.report("parent_scf folder was set")
            parent_calc = self.inputs.parent_scf_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
            if isinstance(parent_calc, PwCalculation):
                if parent_calc.get_state()== 'FINISHED': 
                    if 'settings_pw' not in self.inputs.keys():
                        self.ctx.settings_pw = parent_calc.inp.settings
                    if 'structure' not in self.inputs.keys():
                        self.inputs.structure = parent_calc.inp.structure
                    if 'pseudo' not in self.inputs.keys():
                        raise InputValidationError("Pseudo should be provided")
                    if parent_calc.get_inputs_dict(link_type=LinkType.CREATE)['parameters'].get_dict()['CONTROL']['calculation'] == 'scf':
                        if 'parameters_pw' not in self.inputs.keys():
                            self.ctx.parameters_pw = parent_calc.inp.parameters
                    else:
                        if 'parameters_pw_nscf' not in self.inputs.keys():
                            self.ctx.parameters_pw_nscf = parent_calc.inp.parameters
                    self.report("parent_scf_folder defined params {}".format(self.ctx.parameters_pw.get_dict() ))

            if isinstance(parent_calc, YamboCalculation):
                if parent_calc.get_state()== 'FINISHED': 
                    if 'parameters' not in self.inputs.keys():
                        self.report("setting default parameters, parent was yambocalc ")
                        self.ctx.parameters = set_default_qp_param()

            if 'kpoints'==self.ctx.variable_to_converge:
                if 'settings_pw' not in self.inputs.keys():
                    self.ctx.settings_pw =  default_pw_settings() 
                if 'parameters_pw' not in self.inputs.keys():
                    self.ctx.parameters_pw = set_default_pw_param() 
                if 'parameters_pw_nscf' not in self.inputs.keys():
                    self.ctx.parameters_pw_nscf = set_default_pw_param(nscf=True) 
                      
        else:
            if 'kpoints'==self.ctx.variable_to_converge:
                self.report(" initializing in a kpoints convergence calculation")
                if 'settings_pw' not in self.inputs.keys():
                    self.ctx.settings_pw =  default_pw_settings() 
                if 'parameters_pw' not in self.inputs.keys():
                    self.report("  parameters_pw were not found setting them to default pw params")
                    self.ctx.parameters_pw = set_default_pw_param() 
                if 'parameters_pw_nscf' not in self.inputs.keys():
                    self.report("  parameters_pw_nscf were not found setting them to default pw params")
                    self.ctx.parameters_pw_nscf = set_default_pw_param(nscf=True) 
            if 'kpoints'!=self.ctx.variable_to_converge:
                if 'settings_pw' not in self.inputs.keys():
                    self.ctx.settings_pw =  default_pw_settings() 
                if 'parameters' not in self.inputs.keys():
                    self.ctx.parameters = set_default_qp_param()
            if 'structure' not in self.inputs.keys() :
                raise InputValidationError("Structure should be provided if parent PW SCF folder is not given when converging kpoints")
            if 'pseudo' not in self.inputs.keys():
                raise InputValidationError("Pseudo should be provided if parent PW calculation is not given when converging kpoints")
            if 'pwcode' not in self.inputs.keys():
                raise InputValidationError("PW code  should be provided when converging kpoints")
            if 'kpoints'!=self.ctx.variable_to_converge and\
                                                   'parent_nscf_folder' not in self.inputs.keys():
                raise InputValidationError("Parent nscf folder should be provided when not converging kpoints")

        if 'parent_nscf_folder' in  self.inputs.keys():
              parent_calc = load_node(self.inputs.parent_nscf_folder.get_inputs_dict()['remote_folder'].pk)
              if isinstance(parent_calc, PwCalculation):
                  if parent_calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'nscf'\
                                      and parent_calc.get_state()== 'FINISHED'\
                                      and 'QPkrange' not in self.ctx.parameters.get_dict().keys():
                      self.ctx.parameters = default_qpkrange(parent_calc.pk, self.ctx.parameters)
                      self.report("QPkrange not in  self.ctx.parameters.get_dict keys ")

        if 'settings_pw' in self.inputs.keys():
            self.ctx.settings_pw =  self.inputs.settings_pw 

        params = self.ctx.parameters.get_dict() 
        if 'kpoints'!=self.ctx.variable_to_converge:
            for field in self.ctx.conv_elem.keys():
                 self.report("self.ctx.start_value {}  self.ctx.step {}  self.ctx.loop_length {}  self.ctx.iteration {}".format(
                             self.ctx.start_value, self.ctx.step, self.ctx.loop_length, self.ctx.iteration ))
                 starting_point = self.ctx.start_value +\
                                  self.ctx.step * self.ctx.loop_length * self.ctx.iteration  
                 new_ind_var = update_parameter_field( field, starting_point, self.ctx.step*paging)   
                 if len(self.ctx.conv_elem[field ]) == 1:
                     if self.ctx.conv_elem[field][0] == new_ind_var:
                        pass
                     else:
                        params[ field ] = new_ind_var 
                        self.ctx.conv_elem[field].append(params[field])
                        self.report("initialized {} ".format( params[ field ]  ) )
                 else:
                     params[ field ] = new_ind_var
                     self.ctx.conv_elem[field].append(params[field])
                     self.report("initialized {} ".format( params[ field ]  ) )
            self.ctx.parameters = DataFactory('parameter')(dict= params)
        self.report('Initialization step completed.' )

    def start(self):
        """This function performs some neccessary checks to ensure all neccessary information is provided, and stores  the provided parameters in  self.ctx variables."""
        self.ctx.max_iterations = 20 
        self.ctx.iteration = 0
        self.ctx.skip_prescf = False
        self.ctx.very_first = True
        convergence_parameters_dict = self.inputs.convergence_parameters.get_dict()
        # Mandatory inputs
        try:
            self.ctx.variable_to_converge = convergence_parameters_dict['variable_to_converge']
        except KeyError:
            raise InputValidationError('variable_to_converge not defined in input!')
        try:
            self.ctx.start_value = convergence_parameters_dict['start_value']
        except KeyError:
            raise InputValidationError('start_value not defined in input!')  
        try:
            self.ctx.step = convergence_parameters_dict['step']
        except KeyError:
            raise InputValidationError('step not defined in input!')  
        try:
            self.ctx.max_value = convergence_parameters_dict['max_value']
        except KeyError:            
            raise InputValidationError('max_value not defined in input!') 
        try:                        
            self.ctx.conv_tol = convergence_parameters_dict['conv_tol']
        except KeyError:
            raise InputValidationError('conv_tol not defined in input!')  
        # Optional inputs          
        try:
            self.ctx.conv_window = convergence_parameters_dict['conv_window']
        except KeyError:
            self.ctx.conv_window = 3
        try:
            self.ctx.loop_length = convergence_parameters_dict['loop_length']
        except KeyError:
            self.ctx.loop_length = 4
        self.ctx.distance_kpoints = self.ctx.start_value 
        self.ctx.en_diffs = []
        if self.ctx.variable_to_converge=='bands':
            self.ctx.conv_elem = {'BndsRnXp':[],'GbndRnge': []}
        elif self.ctx.variable_to_converge=='W_cutoff':
            self.ctx.conv_elem = {'NGsBlkXp':[]}
        elif self.ctx.variable_to_converge=='FFT_cutoff':
            self.ctx.conv_elem = {'FFTGvecs':[]}
        elif self.ctx.variable_to_converge=='kpoints':
            self.ctx.conv_elem = {'kpoints':[]}
        else:
            self.ctx.conv_elem = {self.ctx.variable_to_converge:[]}
            self.report('WARNING: the variable to converge is {}, not recognized but I try anyway to converge it') 
        self.report("Setup step completed.")

    def run_next_update(self):
        """This function only stores a completed P2Y, to continued from an NSCF calculation where P2Y was not run, 
        
        This stores a complete P2Y in the ctx to continue from a provided NSCF calculation. 
        """
        if self.ctx.skip_prescf: 
            p2y_parent = load_node( self.ctx.missing_p2y_parent.out.gw.get_dict()["yambo_pk"])
            self.ctx.p2y_parent_folder = p2y_parent.out.remote_folder
            self.ctx.skip_prescf = False

    def iterate(self):
        """This function  creates (and on subsequent iteration updates) the parameters and submits subworkflows  for each parameter step"""
        self.report("Convergence iteration {}".format(str(self.ctx.iteration)))
        loop_items = range(1,self.ctx.loop_length+1)
        if self.ctx.very_first == True:
            loop_items = range(self.ctx.loop_length)
            self.ctx.very_first = False
        self.report('will run four calculations in parallel' )
        outs={}
        if 'kpoints'!=self.ctx.variable_to_converge:
            self.report("this is not a K-point convergence ")
            for num in loop_items: # includes 0 because of starting point
                # There is a bug here, for Bands we might end up 
                # with a situation like BandsRnXP  %   12 | 12  % where both
                # are equal at the very start, this is due to the fact that we need
                # to be able to support BSE calculations too.
                # TODO
                if loop_items[0] == 0:
                    if  'bands' == self.ctx.variable_to_converge:
                        self.init_parameters(num+1) # see above comment
                    else:
                        self.init_parameters(num)
                else:
                    if  'bands' == self.ctx.variable_to_converge:
                        self.update_parameters(num+1)
                    else:
                        self.update_parameters(num)
                try: 
                    p2y_done = self.ctx.p2y_parent_folder 
                except AttributeError:
                    self.report(' no preceeding yambo parent, will run P2Y from NSCF parent first ' )
                    p2y_res =  submit (YamboRestartWf,
                                precode= self.inputs.precode,
                                yambocode=self.inputs.yambocode,
                                parameters = self.inputs.parameters_p2y,
                                calculation_set= self.ctx.calculation_set_p2y,
                                parent_folder = self.inputs.parent_nscf_folder, settings = self.inputs.settings_p2y) 
                    self.ctx.skip_prescf = True
                    #self.ctx.iteration-=1  
                    self.ctx.very_first = True  #  There was a bug  because of this.
                    return ToContext(missing_p2y_parent= p2y_res)
                self.report(' running from preceeding yambo/p2y calculation  ' )
                future =  submit  (YamboRestartWf,
                            precode= self.inputs.precode,
                            yambocode=self.inputs.yambocode,
                            parameters = self.ctx.parameters,
                            calculation_set= self.inputs.calculation_set,
                            parent_folder = self.ctx.p2y_parent_folder, settings = self.inputs.settings)
                outs[ 'r'+str(num) ] =  future
            self.ctx.iteration = self.ctx.iteration + 1
            return ToContext(**outs )
        else:
            # run yambowf, four times. with a different  nscf kpoint starting mesh
            self.report("  K-point convergence commencing")
            mesh = False
            for num in loop_items: # includes 0 because of starting point
                if loop_items[0] == 0:
                    self.init_parameters(num)
                #else:
                #    self.update_parameters(num)  # Not neccessary, kpoint variation is done at PBE level with the self.ctx.distance_kpoints
                def get_kpoints():
                    self.ctx.distance_kpoints = self.ctx.distance_kpoints* 0.9 # 10% change 
                    kpoints = KpointsData()
                    kpoints.set_cell_from_structure(self.inputs.structure)
                    kpoints.set_kpoints_mesh_from_density(distance= self.ctx.distance_kpoints,force_parity=True)
                    return kpoints
                while True:
                    kpoints = get_kpoints()
                    if mesh == kpoints.get_kpoints_mesh(): # deduplicate
                        continue
                    else:
                        break
                mesh =  kpoints.get_kpoints_mesh() 
                extra = {}
                if 'parent_scf_folder' in self.inputs.keys():
                   extra['parent_folder'] = self.inputs.parent_scf_folder
                if 'QPkrange' not in self.ctx.parameters.get_dict().keys():
                   extra['to_set_qpkrange'] = Bool(1)
                if 'BndsRnXp' not in self.ctx.parameters.get_dict().keys() or 'GbndRnge' not in self.ctx.parameters.get_dict().keys() :
                   extra['to_set_bands'] = Bool(1)
                extra['calculation_set_p2y'] = self.ctx.calculation_set_p2y
                extra['calculation_set_pw'] = self.inputs.calculation_set_pw
                extra['settings_p2y'] = self.inputs.settings_p2y
                extra['settings_pw'] = self.ctx.settings_pw
                future =  submit (YamboWorkflow, codename_pw= self.inputs.pwcode, codename_p2y=self.inputs.precode,
                   codename_yambo= self.inputs.yambocode, pseudo_family= self.inputs.pseudo,
                   calculation_set_yambo = self.inputs.calculation_set,
                   settings_yambo=self.inputs.settings , structure = self.inputs.structure,
                   kpoint_pw = kpoints, parameters_pw= self.ctx.parameters_pw, parameters_pw_nscf= self.ctx.parameters_pw_nscf,
                   parameters_p2y= self.inputs.parameters_p2y, parameters_yambo=  self.ctx.parameters,
                   **extra)
                outs[ 'r'+str(num) ] = future
                self.ctx.conv_elem['kpoints'].append(self.ctx.distance_kpoints)
            self.ctx.iteration = self.ctx.iteration + 1
            return ToContext(**outs )  
        return outs 


    def update_parameters(self, paging):
        params = self.ctx.parameters.get_dict()
        for field in self.ctx.conv_elem.keys():
             starting_point = self.ctx.start_value +\
                              self.ctx.step * self.ctx.loop_length* self.ctx.iteration
             if starting_point <0 :
                 self.report("*** ERROR:  updating in the negative direction, please check: starting_point {}".format(starting_point))
                 self.report("*** ERROR:  self.ctx.start_value  {}".format(self.ctx.start_value))
                 self.report("*** ERROR:  self.ctx.step  {}".format(self.ctx.step))
                 self.report("*** ERROR:  self.ctx.loop_length  {}".format(self.ctx.loop_length))
                 self.report("*** ERROR:  self.ctx.iteration  {}".format(self.ctx.iteration))
             params[ field ] = update_parameter_field( field, starting_point, self.ctx.step*paging)
             self.ctx.conv_elem[field].append(params[field])
        self.report(" extended convergence points: {}".format(self.ctx.conv_elem))
        self.ctx.parameters = DataFactory('parameter')(dict= params)


    def is_not_converged(self):
        """This function check is there has been convergence reached or not."""
        if self.ctx.skip_prescf == True:
            return True 
        try: # for yamborestart
            r0_width = self.get_total_range(self.ctx.r1.out.gw.get_dict()['yambo_pk'])
            r1_width = self.get_total_range(self.ctx.r2.out.gw.get_dict()['yambo_pk'])
            r2_width = self.get_total_range(self.ctx.r3.out.gw.get_dict()['yambo_pk'])
            r3_width = self.get_total_range(self.ctx.r4.out.gw.get_dict()['yambo_pk'])
        except AttributeError: # for yamboworkflow
            r0_width = self.get_total_range(self.ctx.r0.out.gw.get_dict()['yambo_pk'])
            r1_width = self.get_total_range(self.ctx.r1.out.gw.get_dict()['yambo_pk'])
            r2_width = self.get_total_range(self.ctx.r2.out.gw.get_dict()['yambo_pk'])
            r3_width = self.get_total_range(self.ctx.r3.out.gw.get_dict()['yambo_pk'])

        self.ctx.en_diffs.extend([r0_width,r1_width,r2_width,r3_width])
        if 'scf_pk' in self.ctx.r1.out.gw.get_dict() and 'parent_scf_folder' not in self.inputs.keys():
            self.inputs.parent_scf_folder =  load_node(self.ctx.r1.out.gw.get_dict()['scf_pk']).out.remote_folder
        if len(self.ctx.en_diffs) > self.ctx.max_iterations: # no more than 16 calcs
            self.report("Aborting after max_iterations={} calculations with no convergence ".format(self.ctx.max_iterations))
            return False
        if 'kpoints' ==self.ctx.variable_to_converge:
            if self.ctx.distance_kpoints <= self.ctx.max_value:
                self.report("Aborting, k-point  convergence not achived within given mesh density range {}- {}".format(0.5,self.ctx.max_value))
                return False
          
        field = self.ctx.conv_elem.keys()[0]
        independent = np.array(self.ctx.conv_elem[field])
        dependent = np.array(self.ctx.en_diffs)
        values = dependent[-4:]
        is_conv =  is_converged(values,conv_tol=self.ctx.conv_tol,conv_window=self.ctx.conv_window)
        if is_conv:
            converged_fit = self.analyse_fit(field)
            if converged_fit:
               self.report("Fit convergence achieved. ")
            else:
               self.report("Fit convergence not achieved.")
            self.report("Absolute  convergence achieved")
            return False
        self.report(" k-point convergence has not been achieved, continuing iterations")
        return True

    def analyse_fit(self,field):
        """This function implements the logic to decide if any >3 calculations have converged.

        arguments:
        field -- The quantity being conveged (kpoint, FFT...)
        independent means the input values.
        dependent the quantity to converge (e.g. band gap).
        This function ensures the energy differences between any four successful calculation to
        be below the threshold, i.e.:
        d1 = C1 - C2
        d2 = C2 - C3
        d3 = C3 - C4
        therefore  d1 and d2 and d3   must be ALL less than the convergence threshold.
        """
        independent = np.array(self.ctx.conv_elem[field])
        if self.ctx.variable_to_converge== 'bands':
            if len(independent.shape) >1:
                independent = independent[:,1]
        dependent = np.array(self.ctx.en_diffs)
        self.report("indep and dependent  {}  {} ".format(independent, dependent))
        def func(x,a,b):
            y = 1.0
            y = y*(a/x+b)
            return y
        popt,pcov = curve_fit(func,independent[-4:],dependent[-4:])
        a,b = popt
        fit_data = func(independent, a,b)      
        deviations = np.abs(dependent[-4:] - fit_data[-4:])  # deviation of last four from extrapolated values at those points
        converged_fit = np.allclose(deviations, np.linspace(0.01,0.01, 4), atol=self.ctx.conv_tol) # last four are within 0.01 of predicted value
        self.report("Fitting_deviation: converged_fit {}".format(converged_fit))
        return converged_fit
 

    def get_total_range(self,node_id):
        """
        compute the gap in energy between two bands along the kpoints listed in QPkrange

        CAVEAT: this does not neccessarily calculate HOMO LUMO gap, but the width between two
                 bands listed in the QPkrange, on the first kpoint selected,
                 i.e. for 'QPkrange': [(1,16,30,31)] , will find width between 
                 kpoint 1  band 30 and kpoint  1 band 31. 
        """
        calc = load_node(node_id)
        table=calc.out.array_qp.get_array('qp_table')
        try:
            qprange = calc.inp.parameters.get_dict()['QPkrange']
        except KeyError: 
            parent_calc =  calc.inp.parent_calc_folder.inp.remote_folder
            if isinstance(parent_calc, YamboCalculation):
                has_found_nelec = False
                while (not has_found_nelec):
                    try:
                        nelec = parent_calc.out.output_parameters.get_dict()['number_of_electrons']
                        has_found_nelec = True
                    except AttributeError:
                        parent_calc = parent_calc.inp.parent_calc_folder.inp.remote_folder
                    except KeyError:
                        parent_calc = parent_calc.inp.parent_calc_folder.inp.remote_folder
            elif isinstance(parent_calc, PwCalculation):
                nelec  = parent_calc.out.output_parameters.get_dict()['number_of_electrons']
            qprange =  [ ( 1, 1 , int(nelec/2) , int(nelec/2)+1 ) ]
        lowest_k = qprange[0][0] # first kpoint listed, 
        lowest_b = qprange[0][-2] # first band on first kpoint listed,
        highest_b= qprange[0][-1]  # last band on first kpoint listed,
        argwlk = np.argwhere(table[:,0]==float(lowest_k))  # indexes for lowest kpoint
        argwlb = np.argwhere(table[:,1]==float(lowest_b))  # indexes for lowest band
        argwhb = np.argwhere(table[:,1]==float(highest_b)) # indexes for highest band
        if len(argwlk)< 1:
            argwlk = np.array([0])
        if len(argwhb) < 1:
            argwhb = np.argwhere(table[:,1]== table[:,1][np.argmax(table[:,1])])
            argwlb = np.argwhere(table[:,1]== table[:,1][np.argmax(table[:,1])]-1 )
        arglb = np.intersect1d(argwlk,argwlb)              # index for lowest kpoints' lowest band
        arghb = np.intersect1d(argwlk,argwhb)              # index for lowest kpoint's highest band
        e_m_eo = calc.out.array_qp.get_array('E_minus_Eo') 
        eo = calc.out.array_qp.get_array('Eo')
        corrected = eo+e_m_eo
        corrected_lb = corrected[arglb]
        corrected_hb = corrected[arghb]
        self.report(" corrected gap(s) {}  at K-point {}, between bands {} and {} for calc {}".format(
                    corrected_hb- corrected_lb, lowest_k, lowest_b, highest_b , node_id))
        return (corrected_hb- corrected_lb)[0]  # for spin polarized there will be two almost equivalent, else just one value.

    def report_wf(self):
        """Output final quantities

        this reports the parameters used in the converged calculation, the pk of that calculation,
        a listing of the space sampled for convergence and the  gaps computed  at those points 
        in the convergence space.
        """
        extra = {}
        nscf_pk = False
        scf_pk = False
        parameters = None
        from aiida.orm import DataFactory
        if 'pw' in self.ctx.r1.out: 
          if 'nscf_pk' in self.ctx.r1.out.pw.get_dict():
              nscf_pk = self.ctx.r1.out.pw.get_dict()['nscf_pk'] 
              self.out("nscf_remote_folder", self.ctx.r1.out.nscf_remote_folder)
          if 'scf_pk' in self.ctx.r1.out.pw.get_dict():
              scf_pk = self.ctx.r1.out.pw.get_dict()['scf_pk'] 
              self.out("scf_remote_folder", self.ctx.r1.out.scf_remote_folder)
        if 'yambo_pk' in self.ctx.r1.out.gw.get_dict():
            parameters = load_node( self.ctx.r1.out.gw.get_dict()['yambo_pk']).inp.parameters.get_dict()
            self.out("yambo_remote_folder", self.ctx.r1.out.yambo_remote_folder)
        self.out("convergence", DataFactory('parameter')(dict={
            "parameters": parameters,
            "yambo_pk": self.ctx.r1.out.gw.get_dict()['yambo_pk'],
            "convergence_space": self.ctx.conv_elem,
            "energy_widths":  self.ctx.en_diffs ,
            "nscf_pk":  nscf_pk, 
            "scf_pk":  scf_pk , 
            }))
        self.report("completed 1-D convergence workflow")

if __name__ == "__main__":
    pass



import sys
from aiida.backends.utils import load_dbenv, is_dbenv_loaded
if not is_dbenv_loaded():
    load_dbenv()

from aiida.common.exceptions import InputValidationError,ValidationError, WorkflowInputValidationError
from aiida.orm import load_node
from aiida.orm.data.base import Float, Str, NumericType, BaseType, List
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida.orm.data.remote import RemoteData
from aiida.common.exceptions import InputValidationError,ValidationError
from aiida.work.run import run, submit
from aiida.work.workchain import WorkChain, while_, ToContext
import numpy as np 
from scipy.optimize import  curve_fit 
from aiida_yambo.workflows.yamboconvergence  import  YamboConvergenceWorkflow
from aiida_yambo.workflows.yambo_utils import default_convergence_settings
from aiida.orm.data.base import Float, Str, NumericType, BaseType, List
from aiida.work.run import run, submit
from aiida.orm.utils import DataFactory
ParameterData = DataFactory("parameter")

class YamboFullConvergenceWorkflow(WorkChain):
    """Full Convergence workflow will converge the direct gap at Gamma  w.r.t. Kpoints, FFT, Bands and G cutoff.

    This convergence workflow follows the following procedure,
    Starts by converging the kpoints, if an SCF calculation is passed in as input, the sCF step will be skipped,
    and computation will begin from the NSCF step, and the  K-point mesh will be varied for the NSCF computation
    that each used fr a yambo GW calculation. This is repeated untill convernence is achived. To optimize for 
    batch queuing systems, the workflow submits 4 different calculations at a time, each representing a full 
    NSCF-GW for a particular mesh size. After each 4 calculations the convergence tested and a decision to move
    to the next parameter or not. 

    After kpoints, the FFT grid is converged, using the k-point mesh from the first step, and similar to the kpoint
    convergence 4 individual GW calculations with different FFT grid sizes are submitted at once, and on completion,
    a test for convergence is performed. This step uses the  NSCF inputs from the converged k-point calculation.

    Following the FFT grid is the bands (BandsRnXP, GbandRnge), which similar to the FFT and kpoints, is performed with four submissions at a
    time followed by a convergence check, using converged k-points and FFT grid from the first two steps.

    Once the Bands converged, we converger the cuttoff for the greens function, which uses the converged parameters from
    the first three steps: k-points, FFT and the Bands. Similar to the preceeding steps,  four GW calculations with different values of cut-off
    are submitted, and convergence is tested. 

    After the cut-off, we redo the bands convergence step, and should the bands converge at a value greater than that from the first bands convergence,
    the cut-off convergence is repeated. This is repeated untill theres is consistency between the bands and cut-off.
    """

    @classmethod
    def define(cls, spec):
        """
        """
        super(YamboFullConvergenceWorkflow, cls).define(spec)

        spec.input("precode", valid_type=BaseType)
        spec.input("pwcode", valid_type=BaseType)
        spec.input("yambocode", valid_type=BaseType)
        spec.input("pseudo", valid_type=BaseType)
        spec.input("threshold", valid_type=Float, required=False, default=Float(0.1))
        spec.input("parent_scf_folder", valid_type=RemoteData, required=False)
        spec.input("structure", valid_type=StructureData,required=False)
        spec.input("calculation_set", valid_type=ParameterData)
        spec.input("calculation_set_pw", valid_type=ParameterData,required=False)
        spec.input("parameters", valid_type=ParameterData,required=False)
        spec.input("parameters_pw", valid_type=ParameterData,required=False)
        spec.input("parameters_pw_nscf", valid_type=ParameterData,required=False)
        spec.input("convergence_settings", valid_type=ParameterData,required=False)
        spec.outline(
          cls.start,
          while_(cls.is_not_converged)(
              cls.run_next_update,
              cls.contex_waits,
              cls.keep_step_data,
              ),
          cls.report_wf
        )
        spec.dynamic_output()

    def init_parameters(self):
        self.ctx.first_run = True
        self.ctx.first_runs2 = True
        self.ctx.first_runs3 = True
        self.ctx.last_step = ''
        self.ctx.step_0_done =False 
        self.ctx.step_1_done =False 
        self.ctx.step_2_done = False 
        self.ctx.step_3_done = False
        self.ctx.bands_n_cutoff_consistent = False
        if 'parent_scf_folder' not in  self.inputs.keys(): 
            self.inputs.parent_scf_folder = False
        if 'structure' not in self.inputs.keys():
            self.inputs.structure = False
        if 'calculation_set_pw' not in self.inputs.keys():
            self.inputs.calculation_set_pw = self.inputs.calculation_set  
        # if input calc has to be SCF not NSCF
        if 'convergence_settings' in self.inputs.keys():
            self.ctx.convergence_settings = self.inputs.convergence_settings
        else:
            self.ctx.convergence_settings = default_convergence_settings()

    def start(self):
        # check that one of structure or parent_scf_calc have been provided.
        if 'parent_scf_folder' not in  self.inputs.keys() and 'structure' not in self.inputs.keys():
           raise InputValidationError("Either the structure or parent SCF calculation should be provided")
        self.ctx.ordered_outputs = [] 
        self.init_parameters()

        """ DEBUG USE ONLY.
        self.ctx.last_step = 'step_1_1'       # select the step  you want to start AFTER.
        self.ctx.step0_res = load_node(37562)  # Fill nodes for prior convergence for all steps you want to skip
        self.ctx.step1_res = load_node(37778)  # "" ""
        self.ctx.step2_res = load_node(45107) # "" ""
        self.ctx.first_run = False            # ""  ""
        self.ctx.scf_calc = self.ctx.step0_res.out.convergence.get_dict()["scf_pk"]     #  ""  "" 
        self.ctx.nscf_calc = self.ctx.step0_res.out.convergence.get_dict()["nscf_pk"]   #  ""  ""
        """
  
    def run_next_update(self):
        """This function will run at each iteration and call the right step depending on what step came before.

        iThe steps are
        step_0 : k-point convergence
        step_1 : FFT convergence
        step_2 : Bands covergence
        step_3 : cutoff convergence
        Each will be called one at a time untill convergence is achieved. 
        See the class `Docstring` to get a description of the algorithm.
        """

        if self.ctx.first_run:
            self.step_0()
        elif self.ctx.last_step == 'step_0_1':
            # Run kpoint  starting with  coarse K-point grid, treat as independent of the rest of the parameters
            self.step_1()
        elif self.ctx.last_step == 'step_1_1':
            # Run FFT from a  coarse FFT grid,  treat as independent, though use converged k-points
            self.step_2()
        elif self.ctx.last_step == 'step_2_1':
            # Run Bands  with  convergged FFT and kpoints
            self.step_3()
        elif self.ctx.last_step == 'step_3_1':
            # Run  G cut-off convergence with converged bands, FFT, kpoints, if different from defaults, else
            # convergence is done 
            self.step_2(recheck=True)
        elif self.ctx.last_step == 'step_2_2':
            # we need to call step_3_2, check if the new converged  bands input value  is
            # different from that used  step_3_1 if so  submit and update last used input
            # else mark self.ctx.bands_n_cutoff_consistent
            # as true
            self.step_3(recheck=True)
        elif self.ctx.last_step == 'step_3_2':
            # We need to call step_2_2 check if the new converged cut-off input value is different
            # from  that of step_3_1, if different, 
            self.step_2(recheck=True)

        if self.ctx.first_run:
            self.ctx.first_run = False

    def contex_waits(self):
        """This function is in the spec.cls,  and is used to resolve the `future` object  and store it in a `self.ctx` variable after each iteration.

        This function will receive the future object stored in a ctx variable by functions that are not in the  spec.cls and can not therefore 
        resolve the `future`  i.e. call `ToContext` on a `future`. These functions will store the subworkflow's  future in the context  and here will
        resolve it and wait for the result by calling ToContext, storing the result in a context variable, which is selected depending on which
        step the  subworkflow represent.

        """
        if self.ctx.last_step == 'step_0_1' or self.ctx.last_step == 'step_0_2':
            return ToContext( step0_res =  self.step0_res_  ) 
        elif self.ctx.last_step == 'step_1_1' or self.ctx.last_step == 'step_1_2':
            return ToContext( step1_res =  self.step1_res_  ) 
        elif self.ctx.last_step == 'step_2_1' or self.ctx.last_step == 'step_2_2':
            return ToContext( step2_res =  self.step2_res_  ) 
        elif self.ctx.last_step == 'step_3_1' or self.ctx.last_step == 'step_3_2':
            return ToContext( step3_res =  self.step3_res_  ) 
        return 

    def keep_step_data(self):
        """Here we store the ordered convergece , SCF and NSCF calcs in the `stx` to keep it available for subsequent calculations that need them, and prevent repetition

        We store the scf and nscf calcs after kpoint convergence, as well as the convergence data after each convergence step in an ordere array to keep
        track of how the workflow evolves to convergence.
        """
        self.report(" persisting outputs to context ")
        if self.ctx.last_step == 'step_0_1' or self.ctx.last_step == 'step_0_2':
            self.ctx.scf_calc = self.ctx.step0_res.out.convergence.get_dict()["scf_pk"]
            self.ctx.nscf_calc = self.ctx.step0_res.out.convergence.get_dict()["nscf_pk"]
            self.report("persisted nscf calc to be used as parent.")

        if self.ctx.last_step == 'step_0_1' or self.ctx.last_step == 'step_0_2':
           self.ctx.ordered_outputs.append( self.ctx.step0_res.out.convergence.get_dict() )
        elif self.ctx.last_step == 'step_1_1' or self.ctx.last_step == 'step_1_2':
           self.ctx.ordered_outputs.append( self.ctx.step1_res.out.convergence.get_dict() )
        elif self.ctx.last_step == 'step_2_1' or self.ctx.last_step == 'step_2_2': # store last conve
           self.ctx.ordered_outputs.append( self.ctx.step2_res.out.convergence.get_dict() )
        elif self.ctx.last_step == 'step_3_1' or  self.ctx.last_step == 'step_3_2':
           self.ctx.ordered_outputs.append( self.ctx.step3_res.out.convergence.get_dict() )
        

    def step_1(self, recheck=False):
        """This calls the YamboConvergenceWorkflow as a subworkflow, converging the FFT grid.

        We converge the FFT grid,  using the converged kpoints from  a preceeding kpoints
        convergence calculation. This is  a 1-D convergence performed by a subworkflow.
        """

        self.report("converging  FFTGvecs")
        extra={}
        if self.inputs.parent_scf_folder:
             extra['parent_scf_folder'] = self.inputs.parent_scf_folder
        if self.inputs.structure:
             extra['structure'] = self.inputs.structure
        if self.ctx.last_step == 'step_2_1':
             extra['parameters'] = ParameterData(dict=self.ctx.step2_res.out.convergence.get_dict()['parameters'] )
        if self.ctx.last_step == 'step_0_1' and self.ctx.step_0_done== True:
             extra['parameters'] = ParameterData(dict=self.ctx.step0_res.out.convergence.get_dict()['parameters'] )
        convergence_parameters = DataFactory('parameter')(dict= { 
                                  'variable_to_converge': 'FFT_cutoff', 'conv_tol':float(self.inputs.threshold), 
                                  'start_value': self.ctx.convergence_settings.dict.start_fft , 'step':20 , # 50
                                  'max_value': self.ctx.convergence_settings.dict.max_fft }) # max 400
        p2y_result = submit(YamboConvergenceWorkflow,
                        pwcode= self.inputs.pwcode,
                        precode= self.inputs.precode ,
                        yambocode=self.inputs.yambocode ,
                        calculation_set= self.inputs.calculation_set ,
                        calculation_set_pw = self.inputs.calculation_set_pw,
                        parent_nscf_folder = load_node(self.ctx.nscf_calc).out.remote_folder, 
                        pseudo = self.inputs.pseudo,
                        convergence_parameters = convergence_parameters, 
                        **extra
                        )
        self.report ("submitted 1-D FFT convergence workflow")
        reslt = {}
        if not recheck :
            reslt['step1_1_res'] = p2y_result 
            self.ctx.last_step = 'step_1_1'
            self.ctx.first_runs1 = False 
            self.ctx.step_1_done = True
        else:
            reslt['step1_2_res'] = p2y_result 
            self.ctx.last_step = 'step_1_2'
         
        self.step1_res_  = p2y_result

    def step_2(self,recheck=False):
        """This calls the YamboConvergenceWorkflow as a subworkflow, converging the  Bands.

        We converge the Bands,  using the converged kpoints from  a preceeding kpoints,
        converged FFT from a preceeding FFT convergence.
        This is  a 1-D convergence performed by a subworkflow.
        """
        self.report ("Working on Bands Convergence ")
        nelec =  load_node(self.ctx.nscf_calc).out.output_parameters.get_dict()['number_of_electrons']
        nbands = load_node(self.ctx.nscf_calc).out.output_parameters.get_dict()['number_of_bands']
        self.ctx.MAX_B_VAL = self.ctx.convergence_settings.dict.max_bands #   int(nelec*8) 
        band_cutoff  = self.ctx.convergence_settings.dict.start_bands #  min(nelec,nbands)
        extra={}
        if self.inputs.parent_scf_folder:
             extra['parent_scf_folder'] = self.inputs.parent_scf_folder
        if self.inputs.structure:
             extra['structure'] = self.inputs.structure
        if self.ctx.last_step == 'step_3_1':
             if self.ctx.step3_res.out.convergence.get_dict()['parameters']['NGsBlkXp'] == self.ctx.convergence_settings.dict.start_w_cutoff  : # 1 == default
                 self.report(" converged cutt-off are similar to the default used in a previous band convergence, consistency achieved")
                 self.ctx.step_2_done = True
                 self.ctx.bands_n_cutoff_consistent = True
                 return 
             #band_cutoff = int(self.ctx.last_used_band *0.7)
             extra['parameters'] = ParameterData(dict=self.ctx.step3_res.out.convergence.get_dict()['parameters'] )
             self.report("updated the bands convergence parameters with cut-off from cutoff convergence step")
        if self.ctx.last_step == 'step_3_2':
             # CRUNCH TIME:  use  values from step3_2
             if self.ctx.step3_res.out.convergence.get_dict()['parameters']['NGsBlkXp'] <=  self.ctx.last_used_cutoff : # 
                 # we are done, and consistent:
                 self.report(" converged cutt-off are similar to those used in a previous bands convergence, consistency achieved")
                 self.ctx.step_2_done = True
                 self.ctx.bands_n_cutoff_consistent = True
                 return 
             self.report("passing parameters from  converged cut-off, this can be repeated untill the two parameters are consistent ")
             #band_cutoff = self.ctx.last_used_band  ## BUG?
             #band_cutoff = int( self.ctx.last_used_band*0.7) 
             extra['parameters'] = ParameterData(dict=self.ctx.step3_res.out.convergence.get_dict()['parameters'] )
             self.report("updated the bands convegence parameters with cut-off from cutoff convergence")
        if self.ctx.last_step != 'step_1_1' and self.ctx.last_step != 'step_1_2': # last iteration was W_cutoff not FFT  
             self.ctx.last_used_cutoff = self.ctx.step3_res.out.convergence.get_dict()['parameters']['NGsBlkXp']
             self.report("Cut-off in last W-Cutoff convergence:  {}".format(self.ctx.last_used_cutoff)) 

        if self.ctx.last_step == 'step_1_1':
             params = self.ctx.step1_res.out.convergence.get_dict()['parameters'] 
             extra['parameters'] = ParameterData(dict=params)
        convergence_parameters = DataFactory('parameter')(dict= { 
                                 'variable_to_converge': 'bands', 'conv_tol':float(self.inputs.threshold),
                                 'start_value': band_cutoff , 'step':10 , #band_cutoff
                                 'max_value': self.ctx.MAX_B_VAL  }) # self.ctx.MAX_B_VAL 
        self.report("converging  BndsRnXp, GbndRnge")

        p2y_result =submit (YamboConvergenceWorkflow,
                        pwcode= self.inputs.pwcode,
                        precode= self.inputs.precode,
                        yambocode=self.inputs.yambocode,
                        calculation_set= self.inputs.calculation_set,
                        calculation_set_pw = self.inputs.calculation_set_pw,
                        parent_nscf_folder = load_node(self.ctx.nscf_calc).out.remote_folder, 
                        pseudo = self.inputs.pseudo,
                        convergence_parameters = convergence_parameters, 
                        #converge_parameters= converge_parameters,
                        #threshold = Float(0.01),starting_points = starting_points, 
                        **extra
                        )
        self.report ("submitted 1-D bands convergence workflow")
        reslt = {}
        if not recheck:
            step2_1_res = p2y_result 
            self.ctx.last_step = 'step_2_1'
            self.ctx.first_runs2 = False 
        else:
            self.step2_2_res = p2y_result 
            self.ctx.last_step = 'step_2_2'
            self.ctx.step_2_done = True
        self.step2_res_ = p2y_result

    def step_3(self, recheck=False):
        """This calls the YamboConvergenceWorkflow as a subworkflow, converging the  G-cutoff.

        We converge the G-cutoff,  using the converged kpoints from  a preceeding kpoints,
        converged FFT from a preceeding FFT convergence and converged Bands from preceeding bands convergence
        This is  a 1-D convergence performed by a subworkflow.
        """
        nbands = load_node(self.ctx.nscf_calc).out.output_parameters.get_dict()['number_of_bands']
        self.report ("Working on W-cutoff ")
        w_cutoff = self.ctx.convergence_settings.dict.start_w_cutoff #2 
        extra={}
        if self.inputs.parent_scf_folder:
             extra['parent_scf_folder'] = self.inputs.parent_scf_folder
        if self.inputs.structure:
             extra['structure'] = self.inputs.structure
        if self.ctx.last_step == 'step_2_2':
             if self.ctx.last_used_band <= self.ctx.step2_res.out.convergence.get_dict()['parameters']['BndsRnXp'][-1] :
                 self.report("bands input is similar to that used in a previous cut-off convergence, consistency achieved") 
                 self.ctx.step_3_done = True
                 self.ctx.bands_n_cutoff_consistent = True
                 return 
             self.report("passing parameters from  re-converged bands ")
             extra['parameters'] = ParameterData(dict=self.ctx.step2_res.out.convergence.get_dict()['parameters'] )
             #w_cutoff =  self.ctx.last_used_cutoff  # start  from last used value. ## BUG?
             #w_cutoff =  int(self.ctx.last_used_cutoff*0.7)  
        if self.ctx.last_step == 'step_2_1':
             self.report("passing parameters from  converged bands ")
             # use cut-off from 2_1
             extra['parameters'] = ParameterData(dict=self.ctx.step2_res.out.convergence.get_dict()['parameters'] )
             #self.ctx.last_used_band = self.ctx.step2_res.out.convergence.get_dict()['parameters']['BndsRnXp'][-1] 
        self.ctx.last_used_band = self.ctx.step2_res.out.convergence.get_dict()['parameters']['BndsRnXp'][-1]
        self.report("Bands in last  bands convergence:  {}".format(self.ctx.last_used_band))
        convergence_parameters = DataFactory('parameter')(dict= { 
                                'variable_to_converge': 'W_cutoff', 'conv_tol':float(self.inputs.threshold), 
                                'start_value': w_cutoff , 'step': 1 ,# w_cutoff
                                'max_value': self.ctx.convergence_settings.dict.max_w_cutoff }) #self.ctx.MAX_B_VAL 
        self.report("converging 1-D  W-off")
        p2y_result = submit(YamboConvergenceWorkflow,
                        pwcode= self.inputs.pwcode,
                        precode= self.inputs.precode ,
                        yambocode=self.inputs.yambocode ,
                        calculation_set= self.inputs.calculation_set ,
                        calculation_set_pw = self.inputs.calculation_set_pw,
                        parent_nscf_folder = load_node(self.ctx.nscf_calc).out.remote_folder, 
                        pseudo = self.inputs.pseudo,
                        convergence_parameters = convergence_parameters,   
                        **extra
                        )
        self.report("Submitted  W-cut off Workflow  ")
        reslt = {}
        if not recheck :
            reslt['step3_1_res'] = p2y_result 
            self.ctx.last_step = 'step_3_1'
            self.ctx.first_runs3 = False 
        else:
            reslt['step3_2_res'] = p2y_result 
            self.ctx.last_step = 'step_3_2'
            self.ctx.step_3_done = True
            # if this  differs from that of step3_1_res 
        self.step3_res_ = p2y_result

    def step_0(self,recheck=False):
        """This calls the YamboConvergenceWorkflow as a subworkflow, converging the K-points.

        We converge the k-points by performing SCF and {NSCF+GW} at different mesh sizes untill convergence.
        This is  a 1-D convergence performed by a subworkflow.
        """
        self.report("Working on K-point convergence ")
        extra={}
        if self.inputs.parent_scf_folder:
             extra['parent_scf_folder'] = self.inputs.parent_scf_folder
        if self.inputs.structure:
             extra['structure'] = self.inputs.structure
        if self.ctx.last_step == 'step_1_1':
             extra['parameters'] = self.ctx.step1_res.out.convergence.get_dict()['parameters'] 
        if 'parameters' in self.inputs.keys():
             extra['parameters'] = self.inputs.parameters
        if 'parameters_pw' in self.inputs.keys():
             extra['parameters_pw'] = self.inputs.parameters_pw
        if 'parameters_pw_nscf' in  self.inputs.keys():
             extra['parameters_pw_nscf'] = self.inputs.parameters_pw_nscf
        self.report("converging K-points ")
        convergence_parameters = DataFactory('parameter')(dict= { 
                                  'variable_to_converge': 'kpoints', 'conv_tol':float(self.inputs.threshold), 
                                  'start_value': self.ctx.convergence_settings.dict.kpoint_starting_distance , 'step':.1, # IGNORE STEP 
                                   'max_value': self.ctx.convergence_settings.dict.kpoint_min_distance }) # 0.34 , 0.0250508117676 
                                   
        p2y_result = submit(YamboConvergenceWorkflow,
                        pwcode= self.inputs.pwcode,
                        precode= self.inputs.precode,
                        yambocode=self.inputs.yambocode,
                        calculation_set= self.inputs.calculation_set,
                        calculation_set_pw = self.inputs.calculation_set_pw,
                        pseudo = self.inputs.pseudo,
                        convergence_parameters = convergence_parameters, 
                        **extra
                        )
        self.report("Submitted the K-point convergence step ")
        reslt = {}
        if not recheck:
            reslt ['step0_1_res'] = p2y_result 
            self.ctx.last_step = 'step_0_1'
            self.ctx.step_0_done = True
        else:
            reslt ['step0_2_res'] = p2y_result 
            self.ctx.last_step = 'step_0_2'
        self.step0_res_ = p2y_result 
        self.ctx.first_run = False
        

    def is_not_converged(self):
        """This function checks if all the individual convergence steps have  been marked as complete, and decides whether the workflow should keep interating with the next step or end calculations when convergence has been achieved."""

        # check we are not complete. 
        if self.ctx.step_0_done and self.ctx.step_1_done and self.ctx.step_2_done and self.ctx.step_3_done and self.ctx.bands_n_cutoff_consistent :
            self.report("convergence reached, workflow will stop")
            return False 
        else:
            self.report("No full  convergence achieved yet")
            self.report("progress:  kpoints converged: {}  , fft converged: {}  , bands converged: {}  ,  cut-off converged:{}".format(
                        self.ctx.step_0_done, self.ctx.step_1_done, self.ctx.step_2_done, self.ctx.step_3_done ))
            self.report("consistency for interdepended bands and cut-off: {}".format(self.ctx.bands_n_cutoff_consistent))
            return True 


    def report_wf(self):
        """Output final quantities"""
        from aiida.orm import DataFactory
        self.out("result", DataFactory('parameter')(dict={
             "kpoints": self.ctx.step0_res.out.convergence.get_dict(),
             "fft": self.ctx.step1_res.out.convergence.get_dict(),
             "bands": self.ctx.step2_res.out.convergence.get_dict(),
             "cutoff": self.ctx.step3_res.out.convergence.get_dict(),
             "ordered_step_output": self.ctx.ordered_outputs
            }))

if __name__ == "__main__":
    pass



# Copyright (C) 2016 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yamboparser
#
#
import os
import re
import numpy as np

#we try to use netcdf
try:
    from netCDF4 import Dataset
except ImportError:
    _has_netcdf = False
else:
    _has_netcdf = True

class YamboFile():
    """
    This is the Yambo file class.
    It takes as input a filename produced by yambo.
    Can be a netcdf or a text file
    
    List of supported NETCDF files:
        -> ndb.QP

    List of supported text files:
        -> r-*_em?1_*_gw0
        -> o-*.qp
    """
    _output_prefixes = ['o-']
    _report_prefixes = ['r-','r.']
    _log_prefixes    = ['l-','l.']
    _p2y_log_prefixes = ['l_']
    _netcdf_prefixes = ['ns','ndb']
    _netcdf_sufixes  = {'QP':'gw','HF_and_locXC':'hf'}

    def __init__(self,filename,folder='.'):
        self.filename = filename
        self.folder   = folder
        self.type     = None   
        self.errors   = [] #list of errors
        self.warnings   = [] #list of warnings
        self.memstats   = [] #list of memory allocation statistics
        self.max_memory = None # max memory allocated or freed (Gb)
        self.last_memory = None # last memory allocated or freed (Gb)
        self.last_memory_time = None # last memory allocated or freed (Gb)
        self.timestats   = [] #list of reported times 
        self.last_time = None # last reported time 
        self.yambo_wrote = None #  yambo performed a write to disk
        self.data     = {} #dictionary containing all the important data from the file
        self.kpoints = {}
        self.timing = []
        self.wall_time = None
        self.game_over = False  # check yambo run completed successfully
        self.p2y_complete = False  # check yambo initialization completed successfully
        self.para_error = False 
        self.unphysical_input = False 
        if any(filename.startswith(prefix) for prefix in self._output_prefixes):
            #read lines from file
            f = open("%s/%s"%(folder,filename),'r')
            self.lines = f.readlines()
            f.close()

            #get the line with the title
            try:
                title = self.lines[14]
            except:
                self.errors.append('error reading title')
                return
            if 'GW' in title:
                 self.type = 'output_gw'

        elif any(filename.startswith(prefix) for prefix in self._report_prefixes):
            self.type = 'report'
        elif any(filename.startswith(prefix) for prefix in self._log_prefixes):
            self.type = 'log'
        elif any(filename.startswith(prefix) for prefix in self._p2y_log_prefixes):
            self.type = 'p2y_log'
        elif any(filename.startswith(prefix) for prefix in self._netcdf_prefixes) and _has_netcdf:
            for sufix in self._netcdf_sufixes:
                if filename.endswith(sufix): 
                    self.type = 'netcdf_%s'%self._netcdf_sufixes[sufix]
                    break

        if self.type is None: self.type = 'unknown'
        
        #parse the file
        self.parse()

    def parse(self):
        """ Parse the file
            Add here things to read log and report files...
        """
        if   self.type == 'netcdf_gw': self.parse_netcdf_gw()
        elif self.type == 'netcdf_hf': self.parse_netcdf_hf()
        elif self.type == 'output_gw': self.parse_output()
        elif self.type == 'log': self.parse_log()
        elif self.type == 'p2y_log': self.parse_p2y_log()
        elif self.type == 'report'  : self.parse_report()

    def parse_output(self):
        """ Parse an output file from yambo,
        """
        #get the tags of the columns
        if self.type == "output_absorption":
            tags = [tag.strip() for tag in re.findall('([ `0-9a-zA-Z\-\/]+)\[[0-9]\]',''.join(self.lines))]
        if self.type == "output_gw":
            tags = [line.replace('(meV)','').replace('Sc(Eo)','Sc|Eo') for line in self.lines if all(tag in line for tag in ['K-point','Band','Eo'])][0]
            tags = tags[2:].strip().split()
        table = np.genfromtxt(self.lines)
        _kdata ={}
        if table.ndim> 1:
            k_index =[ str(int(i)) for i in table[:,0]] # first column  has kpoints
        else:
            k_index =[ str(int(table[0])) ] # first column  has kpoints
        for ind in range(len(k_index)):
            for itag in range(len(tags)):
                 if k_index[ind] not in _kdata.keys():
                     _kdata[k_index[ind]] = {}
                 try:
                     _kdata[k_index[ind]][tags[itag]].append(table[ind,itag])
                 except KeyError:
                    try:
                         _kdata[k_index[ind]][tags[itag]]  = [ table[ind,itag] ]
                    except IndexError:
                         _kdata[k_index[ind]][tags[itag]]  = [ table[itag] ]

        self.data = _kdata 
        #self.data = dict(zip(tags,table.T))

    def parse_netcdf_gw(self):
        """ Parse the netcdf gw file
        """
        if _has_netcdf:
            data = {}
            f = Dataset('%s/%s'%(self.folder,self.filename))
            #quasiparticles table
            qp_table  = f.variables['QP_table'][:].T
            data['Kpoint_index'] = qp_table[:,2]
            data['Band'] = qp_table[:,0]
            if qp_table.shape[1] == 4: # spin polarized
                data['Spin_pol'] = qp_table[:,3]
            data['qp_table'] = qp_table[:,1:]  # ib, ik, ,(isp if spin polarized)
            #qpoints
            data['Kpoint']   = f.variables['QP_kpts'][:].T

            #quasi-particles
            #old format
            if 'QP_E_Eo_Z' in f.variables:
                qp = f.variables['QP_E_Eo_Z'][:]
                qp = qp[0]+qp[1]*1j
                data['E'],  data['Eo'], data['Z'] = qp.T
                data['E-Eo'] = data['E']  -  data['Eo'] 
                self.data=data
                f.close()
            #new format
            else:
                E  = f.variables['QP_E'][:]
                data['E'] = E[:,0] + E[:,1]*1j
                Eo = f.variables['QP_Eo'][:]
                data['Eo']= Eo
                Z  = f.variables['QP_Z'][:]
                data['Z'] = Z[:,0] + Z[:,1]*1j
                data['E-Eo'] = data['E']  -  data['Eo'] 
                self.data=data
                f.close()
       
    def parse_netcdf_hf(self):
        """ Parse the netcdf hf file (ndb.HF_and_locXC)
        """
        if _has_netcdf:
            data = {}
            f = Dataset('%s/%s'%(self.folder,self.filename))
            hf =  f.variables['Sx_Vxc'][:]
            if hf.shape[0]%8 ==0 :
                qp =  hf.reshape(-1,8)
                ib, ibp, ik, isp, rsx, isx, revx, imvx = qp.T
            else:
                qp =  hf.reshape(-1,7)
                ib, ibp, ik, rsx, isx, revx, imvx = qp.T
            data['Sx'] = rsx + isx*1j 
            data['Vxc'] = revx + imvx*1j

            self.data=data
            f.close()
 
    def parse_report(self):
        """ Parse the report files.
            produces output of this nature:
            { k-index1  : { 'dft_enrgy':[...], 'qp_energy':[...] },
              k-index2  :{...}
            }
            k-index is the kpoint at which the yambo calculation was
            done.
        """
        if not hasattr(self, 'lines'):
            with open('%s/%s'%(self.folder,self.filename)) as fl:
                self.lines = fl.readlines()
        # start with check for  failure due to error:
        err = re.compile('^\s+?\[ERROR\]\s+?(.*)$')
        kpoints = re.compile('^  [A-X*]+\sK\s\[([0-9]+)\]\s[:](?:\s+)?([0-9.E-]+\s+[0-9.E-]+\s+[0-9.E-]+)\s[A-Za-z()\s*.]+[0-9]+[A-Za-z()\s*.]+([0-9.]+)')
        timing = re.compile('\s+?[A-Za-z]+iming\s+?[A-Za-z/\[\]]+[:]\s+?([a-z0-9-]+)[/]([a-z0-9-]+)[/]([a-z0-9-]+)')
        game_over = re.compile('^\s+?\[\d+\]\s+?G\w+\s+?O\w+\s+?\&\s+?G\w+\s+?\w+') # Game over & Game summary
        p2y_complete = re.compile('^(\s+)?[-<>\d\w]+\s+?P\d+[:]\s+?==\s+?P2Y\s+?\w+\s+?==(\s+)?') # P2Y Complete
        p2y_complete_v2 = re.compile('(\s+)?[-<>\w\d]+(\s+)?==(\s+)?P2Y(\s+)?\w+(\s+)?==') # P2Y Complete
        yambo_wrote =  re.compile('(?:\s+)?[[]WR[./\w]+[]](?:[-])+')
        for line in self.lines:
            if err.match(line):
                if 'STOP' in err.match(line).groups()[0]:
                    # stop parsing, this is a failed calc.
                    self.errors.append(err.match(line).groups()[0])
                    return 
            if timing.match(line):
                self.timing.append(timing.match(line).groups()[0] )
            if kpoints.match(line):
                kindx, kpt, wgt = kpoints.match(line).groups()
                self.kpoints[str(int(kindx))] =  [ float(i.strip()) for i in kpt.split()]
            if game_over.match(line):
                self.game_over = True
            if p2y_complete.match(line) or p2y_complete_v2.match(line):
                self.p2y_complete = True
                self.game_over = True
            if yambo_wrote.match(line):
                self.yambo_wrote = True
    
        full_lines = ''.join(self.lines)
        qp_regx = re.compile('(^\s+?QP\s\[eV\]\s@\sK\s\[\d+\][a-z0-9E:()\s.-]+)(.*?)(?=^$)',re.M|re.DOTALL)
        kp_regex = re.compile('^\s+?QP\s\[eV\]\s@\sK\s\[(\d+)\][a-z0-9E:()\s.-]+$')
        spliter = re.compile('^(B[=]\d+\sEo[=]\s+?[E0-9.-]+\sE[=]\s+?[E0-9.-]+\sE[-]Eo[=]\s+?[E0-9.-]+\sRe[(]Z[)][=]\s+?[*E0-9.-]+\sIm[(]Z[)][=]\s?[*E0-9.-]+\snlXC[=]\s+?[E0-9.-]+\slXC[=]\s+?[E0-9.-]+\sSo[=]\s+?[E0-9.-]+)')
        extract = re.compile('B[=](\d+)\sEo[=](?:\s+)?([E0-9.-]+)\sE[=](?:\s+)?([E0-9.-]+)\sE[-]Eo[=](?:\s+)?([E0-9.-]+)\sRe[(]Z[)][=](?:\s+)?([*E0-9.-]+)\sIm[(]Z[)][=](?:\s+)?[*E0-9.-]+\snlXC[=](?:\s+)?([E0-9.-]+)\slXC[=](?:\s+)?([E0-9.-]+)\sSo[=](?:\s+)?([E0-9.-]+)')
        qp_lines = qp_regx.findall(full_lines)
        qp_results ={}
        for each in qp_lines: # first group of qp data, shares k-point index
            kp_index = None
            kp_results={'bindex':[],'dft_energy':[],'qp_energy':[],'qp_correction':[],
                        'z_factor':[],'non_local_xc':[],'local_xc':[],'selfenergy_c':[]}
            for line in each: # different band indexes =B
                if kp_regex.match(line):
                    kp_index = str(int(kp_regex.match(line).groups()[0]))
                else: #  data line B=x Eo = y ..
                    data_lines = [ i for i in spliter.split(line) if i.strip()]
                    for qp_data in data_lines:
                        bindex, dft_energy, qp_energy, qp_correction, z_factor, \
                        non_local_xc, local_xc, selfenergy_c = [ i for i in extract.match(qp_data).groups() ]
                        listed = [bindex, dft_energy, qp_energy, qp_correction, z_factor, non_local_xc, local_xc, selfenergy_c]
                        for i in range(len(listed)):
                            try:
                                listed[i] = float(listed[i])
                            except ValueError:
                                listed[i] =  np.nan
                        bindex, dft_energy, qp_energy, qp_correction, z_factor,\
                        non_local_xc, local_xc, selfenergy_c = listed
                        kp_results['bindex'].append(bindex) 
                        kp_results['dft_energy'].append(dft_energy) 
                        kp_results['qp_energy'].append(qp_energy) 
                        kp_results['qp_correction'].append(qp_correction) 
                        kp_results['z_factor'].append(z_factor) 
                        kp_results['non_local_xc'].append(non_local_xc) 
                        kp_results['local_xc'].append(local_xc) 
                        kp_results['selfenergy_c'].append(selfenergy_c) 
            qp_results[kp_index] = kp_results
        self.data = qp_results 

    def get_type(self):
        """ Get the type of file        
        """
        return self.type

    def has_errors(self):
        #check if the list is empty
        return not not self.errors

    def get_errors(self):
        """ Check if this is a report file and if it contains errors
        """
        if self.type == 'report':
            return self.errors
        return False

    def get_data(self):
        """ Get the data from this file as a dictionary 
        """
        pass

    def parse_log(self):
        """ Get ERRORS and WARNINGS from  l-*  file, useful for debugging
        """
        if not hasattr(self, 'lines'):
            with open('%s/%s'%(self.folder,self.filename)) as fl:
                self.lines = fl.readlines()
        warning = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[(WARNING)\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        error = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[(ERROR)\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        self.warnings.extend([ line for line in self.lines if warning.match(line)])
        self.errors.extend([ line for line in self.lines if error.match(line)])
        memory = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[M  ([0-9.]+) Gb\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        self.memstats.extend([ line for line in self.lines if memory.match(line)])
        if len(self.memstats)>0:
            self.max_memory = max([float(memory.match(line).groups()[2]) for line in self.memstats])
            self.last_memory = [float(memory.match(line).groups()[2]) for line in self.memstats][-1]
        # Function to convert time in seconds
        def get_seconds(time_string):
            time = time_string.split('-')
            if time[0]=='' and time[1]=='' and time[2]=='':
                time= ['0h','0m','0s']
            if time[0].endswith('h'):
                return int(time[0].replace("h","")) * 3600 + int(time[1].replace("m","")) * 60 + int(time[2].replace("s",""))
            elif time[0].endswith('m'):
                return int(time[0].replace("m","")) * 60 + int(time[1].replace("s",""))
            elif time[0].endswith('s'):
                return int(time[0].replace("s",""))
            else:
                raise Exception

        times = re.compile('^\s+?<([0-9a-z-]+)>')   
        self.timestats.extend([ get_seconds(times.match(line).groups()[0]) for line in self.lines if times.match(line)]) 
        self.last_time = self.timestats[-1]
        if len(self.memstats)>0:
            self.last_memory_time = [ get_seconds(memory.match(line).groups()[0]) for line in self.memstats][-1]

        generic_error = re.compile('^(?=\s+)?([A-Z0-9]+)[:] \[(ERROR)\](?=\s+)?([a-zA-Z0-9\s.()\[\]]+)?')
        paralle = re.compile('^(?=\s+)?([A-Z0-9]+)[:] \[ERROR\](?=\s+)?Impossible(?=\s+)?(?=[a-zA-Z0-9\s.()\[\]]+)?')
        unphysical = re.compile('^(?=\s+)?([A-Z0-9]+)[:] \[ERROR\](?=\s+)?\[NetCDF\]\s*NetCDF[:]\s*NC_UNLIMITED\s*in\s*the\s*wrong\s*index')
        self.errors.extend ([ line for line in self.lines if ( generic_error.match(line) and (paralle.match(line) or unphysical.match(line)) )  ])
        for line in self.lines:
            if generic_error.match(line):
                if paralle.match(line):
                    self.para_error = True
                if unphysical.match(line):
                    self.unphysical_input = True

    def parse_p2y_log(self):
        """ Get ERRORS and WARNINGS from p2y l_*  file, useful for debugging
            Get *Writing* lines, useful to tell if 'aiida' directory was created.
        """
        if not hasattr(self, 'lines'):
            with open('%s/%s'%(self.folder,self.filename)) as fl:
                self.lines = fl.readlines()
        p2y_complete = re.compile('^(\s+)?[-<>\d\w]+\s+?P\d+[:]\s+?==\s+?P2Y\s+?\w+\s+?==(\s+)?') # P2Y Complete
        p2y_complete_v2 = re.compile('(\s+)?[-<>\w\d]+(\s+)?==(\s+)?P2Y(\s+)?\w+(\s+)?==') # P2Y Complete
        yambo_wrote = re.compile('(?:\s+)?(?:[<>\w\d]+)(:?\s+)?(?:P\d+[:])(?:\s+)?(?:[[]\w+[]])?(?:\s+)?Writing(?:\s+)?\w+(?:\s+)?')
        for line in self.lines:
            if p2y_complete.match(line) or p2y_complete_v2.match(line):
                self.p2y_complete = True
                self.game_over = True 
            if yambo_wrote.match(line):
                self.yambo_wrote = True 

    def __bool__(self):
        if self.type == 'unknown':
            return False
        else:
            return True
    __nonzero__=__bool__

    def __str__(self):
        return "type: %9s   file: %s/%s"%(self.type, self.folder, self.filename)




# Copyright (C) 2016 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yamboparser
# 
from yambofile import *
import os
import numpy as np

class YamboFolder():
    """
    Takes as input a folder name that is the folder where yambo saved r-* o-* l-* and netcdf files
    """

    def __init__(self,path):
        """
        List all the files in the folder and to each of them call YamboFile class
        """
        self.path = path
        self.yambofiles = [] #list of YamboFile instances

        for dirname, dirnames, filenames in os.walk(path):
            # iterate over all the files in the folder
            for filename in filenames:
                y = YamboFile(filename, folder=dirname)
                if y.type !='unknown': #checks if the file is of a known type
                    self.yambofiles.append(y)

    def get_data(self):
        """
        Return a dictionary with all the relevant data in this folder
        """
        data = {}
        for yambofile in self.yambofiles:
            print yambofile.filename
            print "data:",yambofile.data.keys()
            print "memo:",yambofile.memstats
            print "warn:",yambofile.warnings
 
    def __str__(self):
        s = ''
        for yambofile in self.yambofiles:
            s += "%s\n"%str(yambofile)
        return s



import sys
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()

from aiida.orm import load_node
from aiida.common.exceptions import InputValidationError,ValidationError
from aiida.orm.data.upf import get_pseudos_from_structure
from aiida.common.datastructures import calc_states
from collections import defaultdict
from aiida.orm.utils import DataFactory, CalculationFactory

try:
    from aiida.orm.data.base import Float, Str, NumericType, BaseType ,Bool
    from aiida.work.workchain import WorkChain, while_, Outputs
    from aiida.work.workchain import ToContext as ResultToContext
    from aiida.work.run import legacy_workflow
    from aiida.work.run import run, submit
except ImportError:
     pass
from aiida.common.links import LinkType
from aiida.orm.data.remote import RemoteData 
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida_yambo.workflows.yambo_utils import generate_pw_input_params 
from aiida_quantumespresso.calculations.pw import PwCalculation
from aiida_quantumespresso.workflows.pw.base import PwBaseWorkChain
from aiida_yambo.calculations.gw  import YamboCalculation

ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")
#PwCalculation = CalculationFactory('quantumespresso.pw')
PwProcess = PwCalculation.process()
#YamboCalculation = CalculationFactory('yambo.yambo')

class PwRestartWf(WorkChain):
    """This class is a wrapper for the workflows provided by the aiida-quantumespresso plugin

    This class calls the aiida-quantumespresso PW workflow as a subworkflow with the correct
    parameters, and encodes the logic neccessary to decide whether to do an SCF or an NSCF.
    The subworkflow is assumed to include support for restarting failures from recoverable 
    errors such as queue time exhaustion and other similar errors. 
    """

    @classmethod
    def define(cls, spec):
        """Workfunction definition"""
        super(PwRestartWf, cls).define(spec)
        spec.input("codename", valid_type=BaseType)
        spec.input("pseudo_family", valid_type=Str)
        spec.input("calculation_set", valid_type=ParameterData) # custom_scheduler_commands,  resources,...
        spec.input("settings", valid_type=ParameterData)
        spec.input("structure", valid_type=StructureData)
        spec.input("kpoints", valid_type=KpointsData)
        spec.input("gamma", valid_type=Bool, default=Bool(0), required=False)
        spec.input("parameters", valid_type=ParameterData)
        spec.input("parameters_nscf", valid_type=ParameterData ,required=False)
        spec.input("parent_folder", valid_type=RemoteData,required=False)
        spec.outline(
            cls.pwbegin,
            while_(cls.pw_should_continue)(
                cls.pw_continue,
            ),
            cls.report_wf
        )
        spec.dynamic_output()

    def pwbegin(self):
        """This function constructs the correct parameters to be passed to the subworkflow, and will also determine the type (SCF vs NSCF)

        This function evaluates the inputs to  the workflow, and if there is a parent calculation passed, check its  type if SCF, then performs
        and NSCF.  If there is no parent calculation, will default to running an SCF calculation, all done through subworkflows from aiida-quantumespresso
        """
        if 'parent_folder' in  self.inputs.keys() :
            if not isinstance(self.inputs.parent_folder, RemoteData):
                raise InputValidationError("parent_calc_folder when defined must be of"
                                       " type RemoteData")
        parent_folder = None
        inputs={}
        if 'parent_folder' in self.inputs.keys():
            parameters = self.inputs.parameters.get_dict()
            parent_folder=self.inputs.parent_folder
            calc = parent_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
            if calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'scf' and  calc.get_state()== 'FINISHED':# next nscf 
                if  'parameters_nscf' in  self.inputs.keys():
                    parameters = self.inputs.parameters_nscf.get_dict()
                if 'force_symmorphic' not in parameters['SYSTEM']:
                     parameters['SYSTEM']['force_symmorphic'] = True
                if 'nbnd' not in parameters['SYSTEM']:
                     try:
                         parameters['SYSTEM']['nbnd'] = calc.get_outputs_dict()['output_parameters'].get_dict()['number_of_electrons']*20
                     except KeyError:
                         parameters['SYSTEM']['nbnd'] = int(calc.get_outputs_dict()['output_parameters'].get_dict()['number_of_bands']*20) # 
                if 'ELECTRONS' not in parameters:
                   parameters['ELECTRONS'] = {}
                parameters['ELECTRONS']['diagonalization'] = 'davidson'
                parameters['ELECTRONS']['conv_thr'] = 0.000001
                parameters['SYSTEM']['nbnd'] = int(parameters['SYSTEM']['nbnd'])
                parameters['CONTROL']['calculation'] = 'nscf'
                self.report("NSCF PARAMS {}".format(parameters))
                parameters = ParameterData(dict=parameters)
                inputs = generate_pw_input_params(self.inputs.structure, self.inputs.codename, self.inputs.pseudo_family,
                        parameters, self.inputs.calculation_set, self.inputs.kpoints,self.inputs.gamma,self.inputs.settings, parent_folder)
                self.ctx.scf_pk = calc.pk
                self.report(" submitted NSCF {} ".format(calc.pk)) 

            if calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'scf' and  calc.get_state() != 'FINISHED':#  starting from failed SCF
                self.report("restarting failed   SCF  {} ".format(calc.pk))
                inputs = generate_pw_input_params(self.inputs.structure, self.inputs.codename, self.inputs.pseudo_family,
                        self.inputs.parameters, self.inputs.calculation_set, self.inputs.kpoints,self.inputs.gamma,self.inputs.settings, parent_folder)
                
            if calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'nscf' and  calc.get_state()== 'FINISHED':# 
                self.report(" workflow completed nscf successfully, exiting")
                self.ctx.nscf_pk = calc.pk # NSCF is done, we should exit  
        else:
           self.report("Running from SCF")
           inputs = generate_pw_input_params(self.inputs.structure, self.inputs.codename, self.inputs.pseudo_family,
                        self.inputs.parameters, self.inputs.calculation_set, self.inputs.kpoints,self.inputs.gamma,self.inputs.settings, parent_folder)

        future = submit( PwBaseWorkChain, **inputs)
        self.ctx.pw_pks = []
        self.ctx.pw_pks.append(future.pid)
        self.ctx.restart = 0 
        self.ctx.success = False
        self.ctx.scf_pk = None 
        self.ctx.nscf_pk = None
        self.report("submitted subworkflow  {}".format(future.pid))
        return ResultToContext(first_calc=future  )

    def pw_should_continue(self):
        """This function runs after each calculation is finished, and decides whether the calcuations performed was successfull and which one to do next or restart.

        This function encodes the logic for restarting failed  calculations, including deciding if the calculation has failed enough times to warrant exiting the
        workflow. It also decides if the last calculations were successful what calculation to do next, or quit successfully returning computed SCF+NSCF subworkflows.
        """
        if len(self.ctx.pw_pks) ==0: # we never run a single calculation 
            return False 

        if self.ctx.success == True:
            return False
        if self.ctx.restart > 4:
            return False
        if len (self.ctx.pw_pks) < 1: 
            return True 
        calc = None
        if len(self.ctx.pw_pks) ==1:
            calc = load_node(self.ctx.first_calc.out.CALL.pk)
        else:
            calc = load_node(self.ctx.last_calc.out.CALL.pk)
        self.report("calc {} ".format(calc))
        if calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'scf' and  calc.get_state()== 'FINISHED':
            self.ctx.scf_pk = calc.pk 
            self.report(" completed SCF successfully")
            return True 
        if calc.get_state() == calc_states.SUBMISSIONFAILED or calc.get_state() == calc_states.FAILED\
            or 'output_parameters' not in calc.get_outputs_dict()  and  self.ctx.restart < 4:
            self.report(" calculation failed  {}, will try restarting".format(calc.pk))
            return True
        if calc.get_inputs_dict()['parameters'].get_dict()['CONTROL']['calculation'] == 'nscf' and  calc.get_state()== 'FINISHED':
            self.ctx.nscf_pk =  calc.pk 
            self.ctx.success = True
            self.report("completed NSCF successfully, exiting")
            return False 
        if calc.get_state() == calc_states.SUBMISSIONFAILED or calc.get_state() == calc_states.FAILED\
            or 'output_parameters' not in calc.get_outputs_dict()  and  self.ctx.restart >= 4:
            self.ctx.success = False
            self.report("workflow failed to succesfully run any calcultions exiting")
            return False
        self.report("worklfow exiting unsuccessfully ")
        return False

    def pw_continue(self):
        """This function does the actual submissions after the decision has been made whether to continue or not, and will (re)start and SCF (NSCF)

        This function does the actual submission after the first submission done in `self.pwbegin` are complete, and will either restart a failed
        SCF subworkflow or start an NSCF subworkflow from SCF inputs,  or restart a failed NSCF subworkflow
        """
        # restart if neccessary
        calc = None
        if len(self.ctx.pw_pks) ==1:
            calc = load_node(self.ctx.first_calc.out.CALL.pk)
        else:
            calc = load_node(self.ctx.last_calc.out.CALL.pk)
        self.report(" continuing from calculation {}".format(calc.pk))
        parameters = calc.get_inputs_dict()['parameters'].get_dict()
        scf = ''
        parent_folder = None
        if parameters['CONTROL']['calculation'] == 'scf' and calc.get_state()== 'FINISHED':
            scf = 'nscf'
            parent_folder = calc.out.remote_folder
            self.ctx.scf_pk = calc.pk
        elif parameters['CONTROL']['calculation'] == 'scf' and calc.get_state()!= 'FINISHED':
            scf = 'scf'  # RESTART
            parent_folder = calc.out.remote_folder
            self.ctx.restart += 1 # so we do not end up in an infinite loop 
        elif parameters['CONTROL']['calculation'] == 'nscf' and  calc.get_state()== 'FINISHED': 
            self.ctx.nscf_pk = calc.pk
            self.ctx.success = True
            self.ctx.restart += 1 # so we do not end up in an infinite loop 
            return # we are finished, ideally we should not need to arrive here, this is also done at self.pw_should_continue
        elif parameters['CONTROL']['calculation'] == 'nscf' and calc.get_state()!= 'FINISHED':
            scf = 'nscf'  # RESTART
            parent_folder = calc.out.remote_folder
            self.ctx.restart += 1 # so we do not end up in an infinite loop 
        else:
            self.ctx.success = False
            self.report("workflow in an inconsistent state.")
            return 
 
        if 'parameters_nscf' in self.inputs.keys():
            parameters = self.inputs.parameters_nscf.get_dict() 
        if scf == 'nscf':
            if 'force_symmorphic' not in parameters['SYSTEM']:
                 parameters['SYSTEM']['force_symmorphic'] = True 
            if 'nbnd' not in parameters['SYSTEM']:
                 try:
                     parameters['SYSTEM']['nbnd'] = calc.get_outputs_dict()['output_parameters'].get_dict()['number_of_electrons']*20
                 except KeyError:
                     parameters['SYSTEM']['nbnd'] = int(calc.get_outputs_dict()['output_parameters'].get_dict()['number_of_bands']*20) # 
            parameters['SYSTEM']['nbnd'] = int(parameters['SYSTEM']['nbnd'])
        parameters['CONTROL']['calculation'] = scf
        if 'ELECTRONS' not in parameters:
            parameters['ELECTRONS'] = {}
        parameters['ELECTRONS']['diagonalization'] = 'davidson'
        parameters['ELECTRONS']['conv_thr'] = 0.000001
        self.report(" calculation type:  {} and system {}".format(parameters['CONTROL']['calculation'], parameters['SYSTEM'])) 
        parameters = ParameterData(dict=parameters)  
        inputs = generate_pw_input_params(self.inputs.structure, self.inputs.codename, self.inputs.pseudo_family,
                      parameters, self.inputs.calculation_set, self.inputs.kpoints,self.inputs.gamma, self.inputs.settings, parent_folder )
        future =  submit (PwBaseWorkChain, **inputs)
        self.ctx.pw_pks.append(future.pid)
        self.ctx.restart += 1
        self.report("submitted pw  subworkflow  {}".format(future.pid))
        return  ResultToContext(last_calc=future)



    def report_wf(self):
        """Output final quantities

        return information that may be used to figure out
        the status of the calculation.
        """
        self.report("Workflow Complete : scf {}  nscf {} success {}".format(self.ctx.scf_pk,self.ctx.nscf_pk,self.ctx.success))
        from aiida.orm import DataFactory
        res = {}
        if self.ctx.scf_pk:
            res['scf_pk'] = self.ctx.scf_pk
        if self.ctx.nscf_pk:
            res['nscf_pk'] = self.ctx.nscf_pk
        res['success'] = self.ctx.success 
        self.out("pw", DataFactory('parameter')(dict=res )) 
        self.out("scf_remote_folder", load_node(self.ctx.scf_pk).out.remote_folder ) 
        self.out("nscf_remote_folder",  load_node(self.ctx.nscf_pk).out.remote_folder ) 


if __name__ == "__main__":
    pass
